from __future__ import unicode_literals

from spotify import lib, serialized, utils


__all__ = [
    'Error',
    'ErrorType',
    'LibError',
    'Timeout',
]


class Error(Exception):

    """A Spotify error.

    This is the superclass of all custom exceptions raised by pyspotify.
    """

    @classmethod
    def maybe_raise(cls, error_type, ignores=None):
        """Raise an :exc:`LibError` unless the ``error_type`` is
        :attr:`ErrorType.OK` or in the ``ignores`` list of error types.

        Internal method.
        """
        ignores = set(ignores or [])
        ignores.add(ErrorType.OK)
        if error_type not in ignores:
            raise LibError(error_type)


@utils.make_enum('SP_ERROR_')
class ErrorType(utils.IntEnum):
    pass


class LibError(Error):

    """A libspotify error.

    Where many libspotify functions return error codes that must be checked
    after each and every function call, pyspotify raises the
    :exc:`LibError` exception instead. This helps you to not accidentally
    swallow and hide errors when using pyspotify.
    """

    error_type = None
    """The :class:`ErrorType` of the error."""

    @serialized
    def __init__(self, error_type):
        self.error_type = error_type
        message = utils.to_unicode(lib.sp_error_message(error_type))
        super(Error, self).__init__(message)

    def __eq__(self, other):
        return self.error_type == getattr(other, 'error_type', None)

    def __ne__(self, other):
        return not self.__eq__(other)


for attr in dir(lib):
    if attr.startswith('SP_ERROR_'):
        name = attr.replace('SP_ERROR_', '')
        error_no = getattr(lib, attr)
        setattr(LibError, name, LibError(error_no))


class Timeout(Error):

    """Exception raised by an operation not completing within the given
    timeout."""

    def __init__(self, timeout):
        message = 'Operation did not complete in %.3fs' % timeout
        super(Timeout, self).__init__(message)



from __future__ import unicode_literals

import unittest

import spotify
import tests
from tests import mock


@mock.patch('spotify.link.lib', spec=spotify.lib)
class LinkTest(unittest.TestCase):

    def setUp(self):
        self.session = tests.create_session_mock()

    def test_create_without_uri_or_obj_or_sp_link_fails(self, lib_mock):
        with self.assertRaises(AssertionError):
            spotify.Link(self.session)

    def test_create_from_sp_link(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)

        result = spotify.Link(self.session, sp_link=sp_link)

        self.assertEqual(result._sp_link, sp_link)

    def test_create_from_uri(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link

        spotify.Link(self.session, 'spotify:track:foo')

        lib_mock.sp_link_create_from_string.assert_called_once_with(
            mock.ANY)
        self.assertEqual(
            spotify.ffi.string(
                lib_mock.sp_link_create_from_string.call_args[0][0]),
            b'spotify:track:foo')
        self.assertEqual(lib_mock.sp_link_add_ref.call_count, 0)

    def test_create_from_open_spotify_com_url(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link

        spotify.Link(
            self.session,
            'http://open.spotify.com/track/bar ')

        lib_mock.sp_link_create_from_string.assert_called_once_with(
            mock.ANY)
        self.assertEqual(
            spotify.ffi.string(
                lib_mock.sp_link_create_from_string.call_args[0][0]),
            b'spotify:track:bar')

    def test_create_from_play_spotify_com_url(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link

        spotify.Link(
            self.session,
            'https://play.spotify.com/track/bar?gurba=123')

        lib_mock.sp_link_create_from_string.assert_called_once_with(
            mock.ANY)
        self.assertEqual(
            spotify.ffi.string(
                lib_mock.sp_link_create_from_string.call_args[0][0]),
            b'spotify:track:bar')

    def test_raises_error_if_string_isnt_parseable(self, lib_mock):
        lib_mock.sp_link_create_from_string.return_value = spotify.ffi.NULL

        with self.assertRaises(ValueError):
            spotify.Link(self.session, 'invalid link string')

    def test_releases_sp_link_when_link_dies(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link

        link = spotify.Link(self.session, 'spotify:track:foo')
        link = None  # noqa
        tests.gc_collect()

        lib_mock.sp_link_release.assert_called_with(sp_link)

    def test_repr(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        string = 'foo'

        lib_mock.sp_link_as_string.side_effect = tests.buffer_writer(string)
        link = spotify.Link(self.session, string)

        result = repr(link)

        self.assertEqual(result, 'Link(%r)' % string)

    def test_str(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        string = 'foo'

        lib_mock.sp_link_as_string.side_effect = tests.buffer_writer(string)
        link = spotify.Link(self.session, string)

        self.assertEqual(str(link), link.uri)

    def test_eq(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        link1 = spotify.Link(self.session, sp_link=sp_link)
        link2 = spotify.Link(self.session, sp_link=sp_link)

        self.assertTrue(link1 == link2)
        self.assertFalse(link1 == 'foo')

    def test_ne(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        link1 = spotify.Link(self.session, sp_link=sp_link)
        link2 = spotify.Link(self.session, sp_link=sp_link)

        self.assertFalse(link1 != link2)

    def test_hash(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        link1 = spotify.Link(self.session, sp_link=sp_link)
        link2 = spotify.Link(self.session, sp_link=sp_link)

        self.assertEqual(hash(link1), hash(link2))

    def test_uri_grows_buffer_to_fit_link(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        string = 'foo' * 100

        lib_mock.sp_link_as_string.side_effect = tests.buffer_writer(string)
        link = spotify.Link(self.session, string)

        result = link.uri

        lib_mock.sp_link_as_string.assert_called_with(
            sp_link, mock.ANY, mock.ANY)
        self.assertEqual(result, string)

    def test_url_expands_uri_to_http_url(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        string = 'spotify:track:foo'
        lib_mock.sp_link_as_string.side_effect = tests.buffer_writer(string)
        link = spotify.Link(self.session, string)

        result = link.url

        self.assertEqual(result, 'https://open.spotify.com/track/foo')

    def test_type(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_type.return_value = 1
        link = spotify.Link(self.session, 'spotify:track:foo')

        self.assertIs(link.type, spotify.LinkType.TRACK)

        lib_mock.sp_link_type.assert_called_once_with(sp_link)

    @mock.patch('spotify.track.lib', spec=spotify.lib)
    def test_as_track(self, track_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        sp_track = spotify.ffi.cast('sp_track *', 43)
        lib_mock.sp_link_as_track.return_value = sp_track

        link = spotify.Link(self.session, 'spotify:track:foo')
        self.assertEqual(link.as_track()._sp_track, sp_track)

        lib_mock.sp_link_as_track.assert_called_once_with(sp_link)

    @mock.patch('spotify.track.lib', spec=spotify.lib)
    def test_as_track_if_not_a_track(self, track_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_as_track.return_value = spotify.ffi.NULL

        link = spotify.Link(self.session, 'spotify:track:foo')
        self.assertIsNone(link.as_track())

        lib_mock.sp_link_as_track.assert_called_once_with(sp_link)

    @mock.patch('spotify.track.lib', spec=spotify.lib)
    def test_as_track_with_offset(self, track_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        sp_track = spotify.ffi.cast('sp_track *', 43)

        def func(sp_link, offset_ptr):
            offset_ptr[0] = 90
            return sp_track

        lib_mock.sp_link_as_track_and_offset.side_effect = func

        link = spotify.Link(self.session, 'spotify:track:foo')
        offset = link.as_track_offset()

        self.assertEqual(offset, 90)
        lib_mock.sp_link_as_track_and_offset.assert_called_once_with(
            sp_link, mock.ANY)

    @mock.patch('spotify.track.lib', spec=spotify.lib)
    def test_as_track_with_offset_if_not_a_track(
            self, track_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_as_track_and_offset.return_value = spotify.ffi.NULL

        link = spotify.Link(self.session, 'spotify:track:foo')
        offset = link.as_track_offset()

        self.assertIsNone(offset)
        lib_mock.sp_link_as_track_and_offset.assert_called_once_with(
            sp_link, mock.ANY)

    @mock.patch('spotify.album.lib', spec=spotify.lib)
    def test_as_album(self, album_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        sp_album = spotify.ffi.cast('sp_album *', 43)
        lib_mock.sp_link_as_album.return_value = sp_album

        link = spotify.Link(self.session, 'spotify:album:foo')
        self.assertEqual(link.as_album()._sp_album, sp_album)

        lib_mock.sp_link_as_album.assert_called_once_with(sp_link)

    @mock.patch('spotify.album.lib', spec=spotify.lib)
    def test_as_album_if_not_an_album(self, album_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_as_album.return_value = spotify.ffi.NULL

        link = spotify.Link(self.session, 'spotify:album:foo')
        self.assertIsNone(link.as_album())

        lib_mock.sp_link_as_album.assert_called_once_with(sp_link)

    @mock.patch('spotify.artist.lib', spec=spotify.lib)
    def test_as_artist(self, artist_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        sp_artist = spotify.ffi.cast('sp_artist *', 43)
        lib_mock.sp_link_as_artist.return_value = sp_artist

        link = spotify.Link(self.session, 'spotify:artist:foo')
        self.assertEqual(link.as_artist()._sp_artist, sp_artist)

        lib_mock.sp_link_as_artist.assert_called_once_with(sp_link)

    @mock.patch('spotify.artist.lib', spec=spotify.lib)
    def test_as_artist_if_not_an_artist(self, artist_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_as_artist.return_value = spotify.ffi.NULL

        link = spotify.Link(self.session, 'spotify:artist:foo')
        self.assertIsNone(link.as_artist())

        lib_mock.sp_link_as_artist.assert_called_once_with(sp_link)

    @mock.patch('spotify.playlist.lib', spec=spotify.lib)
    def test_as_playlist(self, playlist_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_type.return_value = spotify.LinkType.PLAYLIST
        sp_playlist = spotify.ffi.cast('sp_playlist *', 43)
        lib_mock.sp_playlist_create.return_value = sp_playlist

        link = spotify.Link(self.session, 'spotify:playlist:foo')
        self.assertEqual(link.as_playlist()._sp_playlist, sp_playlist)

        lib_mock.sp_playlist_create.assert_called_once_with(
            self.session._sp_session, sp_link)

        # Since we *created* the sp_playlist, we already have a refcount of 1
        # and shouldn't increase the refcount when wrapping this sp_playlist in
        # an Playlist object
        self.assertEqual(playlist_lib_mock.sp_playlist_add_ref.call_count, 0)

    @mock.patch('spotify.playlist.lib', spec=spotify.lib)
    def test_as_playlist_if_starred(self, playlist_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_type.return_value = spotify.LinkType.STARRED
        sp_playlist = spotify.ffi.cast('sp_playlist *', 43)
        lib_mock.sp_playlist_create.return_value = sp_playlist

        link = spotify.Link(self.session, 'spotify:playlist:foo')
        self.assertEqual(link.as_playlist()._sp_playlist, sp_playlist)

        lib_mock.sp_playlist_create.assert_called_once_with(
            self.session._sp_session, sp_link)

    @mock.patch('spotify.playlist.lib', spec=spotify.lib)
    def test_as_playlist_if_not_a_playlist(self, playlist_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_type.return_value = spotify.LinkType.ARTIST

        link = spotify.Link(self.session, 'spotify:playlist:foo')
        self.assertIsNone(link.as_playlist())

        self.assertEqual(lib_mock.sp_playlist_create.call_count, 0)

    @mock.patch('spotify.user.lib', spec=spotify.lib)
    def test_as_user(self, user_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        sp_user = spotify.ffi.cast('sp_user *', 43)
        lib_mock.sp_link_as_user.return_value = sp_user

        link = spotify.Link(self.session, 'spotify:user:foo')
        self.assertEqual(link.as_user()._sp_user, sp_user)

        lib_mock.sp_link_as_user.assert_called_once_with(sp_link)

    @mock.patch('spotify.user.lib', spec=spotify.lib)
    def test_as_user_if_not_a_user(self, user_lib_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_as_user.return_value = spotify.ffi.NULL

        link = spotify.Link(self.session, 'spotify:user:foo')
        self.assertIsNone(link.as_user())

        lib_mock.sp_link_as_user.assert_called_once_with(sp_link)

    @mock.patch('spotify.Image', spec=spotify.Image)
    def test_as_image(self, image_mock, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_type.return_value = spotify.LinkType.IMAGE
        sp_image = spotify.ffi.cast('sp_image *', 43)
        lib_mock.sp_image_create_from_link.return_value = sp_image
        image_mock.return_value = mock.sentinel.image
        callback = mock.Mock()

        link = spotify.Link(self.session, 'spotify:image:foo')
        result = link.as_image(callback=callback)

        self.assertIs(result, mock.sentinel.image)
        lib_mock.sp_image_create_from_link.assert_called_once_with(
            self.session._sp_session, sp_link)

        # Since we *created* the sp_image, we already have a refcount of 1 and
        # shouldn't increase the refcount when wrapping this sp_image in an
        # Image object
        image_mock.assert_called_with(
            self.session, sp_image=sp_image, add_ref=False, callback=callback)

    def test_as_image_if_not_a_image(self, lib_mock):
        sp_link = spotify.ffi.cast('sp_link *', 42)
        lib_mock.sp_link_create_from_string.return_value = sp_link
        lib_mock.sp_link_type.return_value = spotify.LinkType.ARTIST

        link = spotify.Link(self.session, 'spotify:image:foo')
        result = link.as_image()

        self.assertIsNone(result)
        self.assertEqual(lib_mock.sp_image_create_from_link.call_count, 0)


class LinkTypeTest(unittest.TestCase):

    def test_has_link_type_constants(self):
        self.assertEqual(spotify.LinkType.INVALID, 0)
        self.assertEqual(spotify.LinkType.TRACK, 1)
        self.assertEqual(spotify.LinkType.ALBUM, 2)



# encoding: utf-8

from __future__ import unicode_literals

import unittest

import spotify
from spotify.session import _SessionCallbacks

import tests
from tests import mock


@mock.patch('spotify.session.lib', spec=spotify.lib)
class SessionTest(unittest.TestCase):

    def tearDown(self):
        spotify._session_instance = None

    def test_raises_error_if_a_session_already_exists(self, lib_mock):
        tests.create_real_session(lib_mock)

        with self.assertRaises(RuntimeError):
            tests.create_real_session(lib_mock)

    @mock.patch('spotify.Config')
    def test_creates_config_if_none_provided(self, config_cls_mock, lib_mock):
        lib_mock.sp_session_create.return_value = spotify.ErrorType.OK

        session = spotify.Session()

        config_cls_mock.assert_called_once_with()
        self.assertEqual(session.config, config_cls_mock.return_value)

    @mock.patch('spotify.Config')
    def test_tries_to_load_application_key_if_none_provided(
            self, config_cls_mock, lib_mock):
        lib_mock.sp_session_create.return_value = spotify.ErrorType.OK
        config_mock = config_cls_mock.return_value
        config_mock.application_key = None

        spotify.Session()

        config_mock.load_application_key_file.assert_called_once_with()

    def test_raises_error_if_not_ok(self, lib_mock):
        lib_mock.sp_session_create.return_value = (
            spotify.ErrorType.BAD_API_VERSION)
        config = spotify.Config()
        config.application_key = b'\x01' * 321

        with self.assertRaises(spotify.Error):
            spotify.Session(config=config)

    def test_releases_sp_session_when_session_dies(self, lib_mock):
        sp_session = spotify.ffi.NULL

        def func(sp_session_config, sp_session_ptr):
            sp_session_ptr[0] = sp_session
            return spotify.ErrorType.OK

        lib_mock.sp_session_create.side_effect = func
        config = spotify.Config()
        config.application_key = b'\x01' * 321

        session = spotify.Session(config=config)
        session = None  # noqa
        spotify._session_instance = None
        tests.gc_collect()

        lib_mock.sp_session_release.assert_called_with(sp_session)

    def test_login_raises_error_if_no_password_and_no_blob(self, lib_mock):
        lib_mock.sp_session_login.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(AttributeError):
            session.login('alice')

    def test_login_with_password(self, lib_mock):
        lib_mock.sp_session_login.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        session.login('alice', 'secret')

        lib_mock.sp_session_login.assert_called_once_with(
            session._sp_session, mock.ANY, mock.ANY,
            False, spotify.ffi.NULL)
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_session_login.call_args[0][1]),
            b'alice')
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_session_login.call_args[0][2]),
            b'secret')

    def test_login_with_blob(self, lib_mock):
        lib_mock.sp_session_login.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        session.login('alice', blob='secret blob')

        lib_mock.sp_session_login.assert_called_once_with(
            session._sp_session, mock.ANY, spotify.ffi.NULL,
            False, mock.ANY)
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_session_login.call_args[0][1]),
            b'alice')
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_session_login.call_args[0][4]),
            b'secret blob')

    def test_login_with_remember_me_flag(self, lib_mock):
        lib_mock.sp_session_login.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        session.login('alice', 'secret', remember_me='anything truish')

        lib_mock.sp_session_login.assert_called_once_with(
            session._sp_session, mock.ANY, mock.ANY,
            True, spotify.ffi.NULL)

    def test_login_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_login.return_value = spotify.ErrorType.NO_SUCH_USER
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.login('alice', 'secret')

    def test_logout(self, lib_mock):
        lib_mock.sp_session_logout.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        session.logout()

        lib_mock.sp_session_logout.assert_called_once_with(session._sp_session)

    def test_logout_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_login.return_value = (
            spotify.ErrorType.BAD_API_VERSION)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.logout()

    def test_remembered_user_name_grows_buffer_to_fit_username(self, lib_mock):
        username = 'alice' * 100

        lib_mock.sp_session_remembered_user.side_effect = (
            tests.buffer_writer(username))
        session = tests.create_real_session(lib_mock)

        result = session.remembered_user_name

        lib_mock.sp_session_remembered_user.assert_called_with(
            session._sp_session, mock.ANY, mock.ANY)
        self.assertEqual(result, username)

    def test_remembered_user_name_is_none_if_not_remembered(self, lib_mock):
        lib_mock.sp_session_remembered_user.return_value = -1
        session = tests.create_real_session(lib_mock)

        result = session.remembered_user_name

        lib_mock.sp_session_remembered_user.assert_called_with(
            session._sp_session, mock.ANY, mock.ANY)
        self.assertIsNone(result)

    def test_relogin(self, lib_mock):
        lib_mock.sp_session_relogin.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        session.relogin()

        lib_mock.sp_session_relogin.assert_called_once_with(
            session._sp_session)

    def test_relogin_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_relogin.return_value = (
            spotify.ErrorType.NO_CREDENTIALS)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.relogin()

    def test_forget_me(self, lib_mock):
        lib_mock.sp_session_forget_me.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        session.forget_me()

        lib_mock.sp_session_forget_me.assert_called_with(session._sp_session)

    def test_forget_me_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_forget_me.return_value = (
            spotify.ErrorType.BAD_API_VERSION)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.forget_me()

    @mock.patch('spotify.user.lib', spec=spotify.lib)
    def test_user(self, user_lib_mock, lib_mock):
        lib_mock.sp_session_user.return_value = (
            spotify.ffi.cast('sp_user *', 42))
        session = tests.create_real_session(lib_mock)

        result = session.user

        lib_mock.sp_session_user.assert_called_with(session._sp_session)
        self.assertIsInstance(result, spotify.User)

    def test_user_if_not_logged_in(self, lib_mock):
        lib_mock.sp_session_user.return_value = spotify.ffi.NULL
        session = tests.create_real_session(lib_mock)

        result = session.user

        lib_mock.sp_session_user.assert_called_with(session._sp_session)
        self.assertIsNone(result)

    def test_user_name(self, lib_mock):
        lib_mock.sp_session_user_name.return_value = spotify.ffi.new(
            'char[]', b'alice')
        session = tests.create_real_session(lib_mock)

        result = session.user_name

        lib_mock.sp_session_user_name.assert_called_with(session._sp_session)
        self.assertEqual(result, 'alice')

    def test_user_country(self, lib_mock):
        lib_mock.sp_session_user_country.return_value = (
            ord('S') << 8 | ord('E'))
        session = tests.create_real_session(lib_mock)

        result = session.user_country

        lib_mock.sp_session_user_country.assert_called_with(
            session._sp_session)
        self.assertEqual(result, 'SE')

    @mock.patch('spotify.playlist_container.lib', spec=spotify.lib)
    def test_playlist_container(self, playlist_lib_mock, lib_mock):
        lib_mock.sp_session_playlistcontainer.return_value = (
            spotify.ffi.cast('sp_playlistcontainer *', 42))
        session = tests.create_real_session(lib_mock)

        result = session.playlist_container

        lib_mock.sp_session_playlistcontainer.assert_called_with(
            session._sp_session)
        self.assertIsInstance(result, spotify.PlaylistContainer)

    @mock.patch('spotify.playlist_container.lib', spec=spotify.lib)
    def test_playlist_container_if_already_listened_to(
            self, playlist_lib_mock, lib_mock):
        lib_mock.sp_session_playlistcontainer.return_value = (
            spotify.ffi.cast('sp_playlistcontainer *', 42))
        session = tests.create_real_session(lib_mock)

        result1 = session.playlist_container
        result1.on(
            spotify.PlaylistContainerEvent.PLAYLIST_ADDED, lambda *args: None)
        result2 = session.playlist_container

        result1.off()

        self.assertIsInstance(result1, spotify.PlaylistContainer)
        self.assertIs(result1, result2)

    def test_playlist_container_if_not_logged_in(self, lib_mock):
        lib_mock.sp_session_playlistcontainer.return_value = spotify.ffi.NULL
        session = tests.create_real_session(lib_mock)

        result = session.playlist_container

        lib_mock.sp_session_playlistcontainer.assert_called_with(
            session._sp_session)
        self.assertIsNone(result)

    @mock.patch('spotify.playlist.lib', spec=spotify.lib)
    def test_inbox(self, playlist_lib_mock, lib_mock):
        lib_mock.sp_session_inbox_create.return_value = (
            spotify.ffi.cast('sp_playlist *', 42))
        session = tests.create_real_session(lib_mock)

        result = session.inbox

        lib_mock.sp_session_inbox_create.assert_called_with(
            session._sp_session)
        self.assertIsInstance(result, spotify.Playlist)

        # Since we *created* the sp_playlist, we already have a refcount of 1
        # and shouldn't increase the refcount when wrapping this sp_playlist in
        # a Playlist object
        self.assertEqual(playlist_lib_mock.sp_playlist_add_ref.call_count, 0)

    def test_inbox_if_not_logged_in(self, lib_mock):
        lib_mock.sp_session_inbox_create.return_value = spotify.ffi.NULL
        session = tests.create_real_session(lib_mock)

        result = session.inbox

        lib_mock.sp_session_inbox_create.assert_called_with(
            session._sp_session)
        self.assertIsNone(result)

    def test_set_cache_size(self, lib_mock):
        lib_mock.sp_session_set_cache_size.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        session.set_cache_size(100)

        lib_mock.sp_session_set_cache_size.assert_called_once_with(
            session._sp_session, 100)

    def test_set_cache_size_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_set_cache_size.return_value = (
            spotify.ErrorType.BAD_API_VERSION)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.set_cache_size(100)

    def test_flush_caches(self, lib_mock):
        lib_mock.sp_session_flush_caches.return_value = spotify.ErrorType.OK
        session = tests.create_real_session(lib_mock)

        session.flush_caches()

        lib_mock.sp_session_flush_caches.assert_called_once_with(
            session._sp_session)

    def test_flush_caches_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_flush_caches.return_value = (
            spotify.ErrorType.BAD_API_VERSION)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.flush_caches()

    def test_preferred_bitrate(self, lib_mock):
        lib_mock.sp_session_preferred_bitrate.return_value = (
            spotify.ErrorType.OK)
        session = tests.create_real_session(lib_mock)

        session.preferred_bitrate(spotify.Bitrate.BITRATE_320k)

        lib_mock.sp_session_preferred_bitrate.assert_called_with(
            session._sp_session, spotify.Bitrate.BITRATE_320k)

    def test_preferred_bitrate_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_preferred_bitrate.return_value = (
            spotify.ErrorType.INVALID_ARGUMENT)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.preferred_bitrate(17)

    def test_preferred_offline_bitrate(self, lib_mock):
        lib_mock.sp_session_preferred_offline_bitrate.return_value = (
            spotify.ErrorType.OK)
        session = tests.create_real_session(lib_mock)

        session.preferred_offline_bitrate(spotify.Bitrate.BITRATE_320k)

        lib_mock.sp_session_preferred_offline_bitrate.assert_called_with(
            session._sp_session, spotify.Bitrate.BITRATE_320k, 0)

    def test_preferred_offline_bitrate_with_allow_resync(self, lib_mock):
        lib_mock.sp_session_preferred_offline_bitrate.return_value = (
            spotify.ErrorType.OK)
        session = tests.create_real_session(lib_mock)

        session.preferred_offline_bitrate(
            spotify.Bitrate.BITRATE_320k, allow_resync=True)

        lib_mock.sp_session_preferred_offline_bitrate.assert_called_with(
            session._sp_session, spotify.Bitrate.BITRATE_320k, 1)

    def test_preferred_offline_bitrate_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_preferred_offline_bitrate.return_value = (
            spotify.ErrorType.INVALID_ARGUMENT)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.preferred_offline_bitrate(17)

    def test_get_volume_normalization(self, lib_mock):
        lib_mock.sp_session_get_volume_normalization.return_value = 0
        session = tests.create_real_session(lib_mock)

        result = session.volume_normalization

        lib_mock.sp_session_get_volume_normalization.assert_called_with(
            session._sp_session)
        self.assertFalse(result)

    def test_set_volume_normalization(self, lib_mock):
        lib_mock.sp_session_set_volume_normalization.return_value = (
            spotify.ErrorType.OK)
        session = tests.create_real_session(lib_mock)

        session.volume_normalization = True

        lib_mock.sp_session_set_volume_normalization.assert_called_with(
            session._sp_session, 1)

    def test_set_volume_normalization_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_set_volume_normalization.return_value = (
            spotify.ErrorType.BAD_API_VERSION)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.volume_normalization = True

    def test_process_events_returns_ms_to_next_timeout(self, lib_mock):
        def func(sp_session, int_ptr):
            int_ptr[0] = 5500
            return spotify.ErrorType.OK

        lib_mock.sp_session_process_events.side_effect = func

        session = tests.create_real_session(lib_mock)

        timeout = session.process_events()

        self.assertEqual(timeout, 5500)

    def test_process_events_fail_raises_error(self, lib_mock):
        lib_mock.sp_session_process_events.return_value = (
            spotify.ErrorType.BAD_API_VERSION)
        session = tests.create_real_session(lib_mock)

        with self.assertRaises(spotify.Error):
            session.process_events()

    @mock.patch('spotify.InboxPostResult', spec=spotify.InboxPostResult)
    def test_inbox_post_tracks(self, inbox_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        inbox_instance_mock = inbox_mock.return_value

        result = session.inbox_post_tracks(
            mock.sentinel.username, mock.sentinel.tracks,
            mock.sentinel.message, mock.sentinel.callback)

        inbox_mock.assert_called_with(
            session, mock.sentinel.username, mock.sentinel.tracks,
            mock.sentinel.message, mock.sentinel.callback)
        self.assertEqual(result, inbox_instance_mock)

    @mock.patch('spotify.playlist.lib', spec=spotify.lib)
    def test_get_starred(self, playlist_lib_mock, lib_mock):
        lib_mock.sp_session_starred_for_user_create.return_value = (
            spotify.ffi.cast('sp_playlist *', 42))
        session = tests.create_real_session(lib_mock)

        result = session.get_starred('alice')

        lib_mock.sp_session_starred_for_user_create.assert_called_with(
            session._sp_session, b'alice')
        self.assertIsInstance(result, spotify.Playlist)

        # Since we *created* the sp_playlist, we already have a refcount of 1
        # and shouldn't increase the refcount when wrapping this sp_playlist in
        # a Playlist object
        self.assertEqual(playlist_lib_mock.sp_playlist_add_ref.call_count, 0)

    @mock.patch('spotify.playlist.lib', spec=spotify.lib)
    def test_get_starred_for_current_user(self, playlist_lib_mock, lib_mock):
        lib_mock.sp_session_starred_create.return_value = (
            spotify.ffi.cast('sp_playlist *', 42))
        session = tests.create_real_session(lib_mock)

        result = session.get_starred()

        lib_mock.sp_session_starred_create.assert_called_with(
            session._sp_session)
        self.assertIsInstance(result, spotify.Playlist)

        # Since we *created* the sp_playlist, we already have a refcount of 1
        # and shouldn't increase the refcount when wrapping this sp_playlist in
        # a Playlist object
        self.assertEqual(playlist_lib_mock.sp_playlist_add_ref.call_count, 0)

    def test_get_starred_if_not_logged_in(self, lib_mock):
        lib_mock.sp_session_starred_for_user_create.return_value = (
            spotify.ffi.NULL)
        session = tests.create_real_session(lib_mock)

        result = session.get_starred('alice')

        lib_mock.sp_session_starred_for_user_create.assert_called_with(
            session._sp_session, b'alice')
        self.assertIsNone(result)

    @mock.patch('spotify.playlist_container.lib', spec=spotify.lib)
    def test_get_published_playlists(self, playlist_lib_mock, lib_mock):
        func_mock = lib_mock.sp_session_publishedcontainer_for_user_create
        func_mock.return_value = spotify.ffi.cast('sp_playlistcontainer *', 42)
        session = tests.create_real_session(lib_mock)

        result = session.get_published_playlists('alice')

        func_mock.assert_called_with(session._sp_session, b'alice')
        self.assertIsInstance(result, spotify.PlaylistContainer)

        # Since we *created* the sp_playlistcontainer, we already have a
        # refcount of 1 and shouldn't increase the refcount when wrapping this
        # sp_playlistcontainer in a PlaylistContainer object
        self.assertEqual(
            playlist_lib_mock.sp_playlistcontainer_add_ref.call_count, 0)

    @mock.patch('spotify.playlist_container.lib', spec=spotify.lib)
    def test_get_published_playlists_for_current_user(
            self, playlist_lib_mock, lib_mock):
        func_mock = lib_mock.sp_session_publishedcontainer_for_user_create
        func_mock.return_value = spotify.ffi.cast('sp_playlistcontainer *', 42)
        session = tests.create_real_session(lib_mock)

        result = session.get_published_playlists()

        func_mock.assert_called_with(session._sp_session, spotify.ffi.NULL)
        self.assertIsInstance(result, spotify.PlaylistContainer)

    def test_get_published_playlists_if_not_logged_in(self, lib_mock):
        func_mock = lib_mock.sp_session_publishedcontainer_for_user_create
        func_mock.return_value = spotify.ffi.NULL
        session = tests.create_real_session(lib_mock)

        result = session.get_published_playlists('alice')

        func_mock.assert_called_with(session._sp_session, b'alice')
        self.assertIsNone(result)

    @mock.patch('spotify.Link')
    def test_get_link(self, link_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        link_mock.return_value = mock.sentinel.link

        result = session.get_link('spotify:any:foo')

        self.assertIs(result, mock.sentinel.link)
        link_mock.assert_called_with(session, uri='spotify:any:foo')

    @mock.patch('spotify.Track')
    def test_get_track(self, track_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        track_mock.return_value = mock.sentinel.track

        result = session.get_track('spotify:track:foo')

        self.assertIs(result, mock.sentinel.track)
        track_mock.assert_called_with(session, uri='spotify:track:foo')

    @mock.patch('spotify.Track')
    def test_get_local_track(self, track_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        sp_track = spotify.ffi.cast('sp_track *', 42)
        lib_mock.sp_localtrack_create.return_value = sp_track
        track_mock.return_value = mock.sentinel.track

        track = session.get_local_track(
            artist='foo', title='bar', album='baz', length=210000)

        self.assertEqual(track, mock.sentinel.track)
        lib_mock.sp_localtrack_create.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY, 210000)
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_localtrack_create.call_args[0][0]),
            b'foo')
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_localtrack_create.call_args[0][1]),
            b'bar')
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_localtrack_create.call_args[0][2]),
            b'baz')
        self.assertEqual(
            lib_mock.sp_localtrack_create.call_args[0][3], 210000)

        # Since we *created* the sp_track, we already have a refcount of 1 and
        # shouldn't increase the refcount when wrapping this sp_track in a
        # Track object
        track_mock.assert_called_with(
            session, sp_track=sp_track, add_ref=False)

    @mock.patch('spotify.Track')
    def test_get_local_track_with_defaults(self, track_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        sp_track = spotify.ffi.cast('sp_track *', 42)
        lib_mock.sp_localtrack_create.return_value = sp_track
        track_mock.return_value = mock.sentinel.track

        track = session.get_local_track()

        self.assertEqual(track, mock.sentinel.track)
        lib_mock.sp_localtrack_create.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY, -1)
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_localtrack_create.call_args[0][0]),
            b'')
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_localtrack_create.call_args[0][1]),
            b'')
        self.assertEqual(
            spotify.ffi.string(lib_mock.sp_localtrack_create.call_args[0][2]),
            b'')
        self.assertEqual(
            lib_mock.sp_localtrack_create.call_args[0][3], -1)

        # Since we *created* the sp_track, we already have a refcount of 1 and
        # shouldn't increase the refcount when wrapping this sp_track in a
        # Track object
        track_mock.assert_called_with(
            session, sp_track=sp_track, add_ref=False)

    @mock.patch('spotify.Album')
    def test_get_album(self, album_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        album_mock.return_value = mock.sentinel.album

        result = session.get_album('spotify:album:foo')

        self.assertIs(result, mock.sentinel.album)
        album_mock.assert_called_with(session, uri='spotify:album:foo')

    @mock.patch('spotify.Artist')
    def test_get_artist(self, artist_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        artist_mock.return_value = mock.sentinel.artist

        result = session.get_artist('spotify:artist:foo')

        self.assertIs(result, mock.sentinel.artist)
        artist_mock.assert_called_with(session, uri='spotify:artist:foo')

    @mock.patch('spotify.Playlist')
    def test_get_playlist(self, playlist_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        playlist_mock.return_value = mock.sentinel.playlist

        result = session.get_playlist('spotify:playlist:foo')

        self.assertIs(result, mock.sentinel.playlist)
        playlist_mock.assert_called_with(session, uri='spotify:playlist:foo')

    @mock.patch('spotify.User')
    def test_get_user(self, user_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        user_mock.return_value = mock.sentinel.user

        result = session.get_user('spotify:user:foo')

        self.assertIs(result, mock.sentinel.user)
        user_mock.assert_called_with(session, uri='spotify:user:foo')

    @mock.patch('spotify.Image')
    def test_get_image(self, image_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        callback = mock.Mock()
        image_mock.return_value = mock.sentinel.image

        result = session.get_image('spotify:image:foo', callback=callback)

        self.assertIs(result, mock.sentinel.image)
        image_mock.assert_called_with(
            session, uri='spotify:image:foo', callback=callback)

    @mock.patch('spotify.Search')
    def test_search(self, search_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        search_mock.return_value = mock.sentinel.search

        result = session.search('alice')

        self.assertIs(result, mock.sentinel.search)
        search_mock.assert_called_with(
            session, query='alice', callback=None,
            track_offset=0, track_count=20,
            album_offset=0, album_count=20,
            artist_offset=0, artist_count=20,
            playlist_offset=0, playlist_count=20,
            search_type=None)

    @mock.patch('spotify.Toplist')
    def test_toplist(self, toplist_mock, lib_mock):
        session = tests.create_real_session(lib_mock)
        toplist_mock.return_value = mock.sentinel.toplist

        result = session.get_toplist(
            type=spotify.ToplistType.TRACKS, region='NO')

        self.assertIs(result, mock.sentinel.toplist)
        toplist_mock.assert_called_with(
            session, type=spotify.ToplistType.TRACKS, region='NO',
            canonical_username=None, callback=None)


@mock.patch('spotify.session.lib', spec=spotify.lib)
class SessionCallbacksTest(unittest.TestCase):

    def tearDown(self):
        spotify._session_instance = None

    def test_logged_in_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.LOGGED_IN, callback)

        _SessionCallbacks.logged_in(
            session._sp_session, int(spotify.ErrorType.BAD_API_VERSION))

        callback.assert_called_once_with(
            session, spotify.ErrorType.BAD_API_VERSION)

    def test_logged_out_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.LOGGED_OUT, callback)

        _SessionCallbacks.logged_out(session._sp_session)

        callback.assert_called_once_with(session)

    def test_metadata_updated_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.METADATA_UPDATED, callback)

        _SessionCallbacks.metadata_updated(session._sp_session)

        callback.assert_called_once_with(session)

    def test_connection_error_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.CONNECTION_ERROR, callback)

        _SessionCallbacks.connection_error(
            session._sp_session, int(spotify.ErrorType.OK))

        callback.assert_called_once_with(session, spotify.ErrorType.OK)

    def test_message_to_user_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.MESSAGE_TO_USER, callback)
        data = spotify.ffi.new('char[]', b'a log message\n')

        _SessionCallbacks.message_to_user(session._sp_session, data)

        callback.assert_called_once_with(session, 'a log message')

    def test_notify_main_thread_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.NOTIFY_MAIN_THREAD, callback)

        _SessionCallbacks.notify_main_thread(session._sp_session)

        callback.assert_called_once_with(session)

    def test_music_delivery_callback(self, lib_mock):
        sp_audioformat = spotify.ffi.new('sp_audioformat *')
        sp_audioformat.channels = 2
        audio_format = spotify.AudioFormat(sp_audioformat)

        num_frames = 10
        frames_size = audio_format.frame_size() * num_frames
        frames = spotify.ffi.new('char[]', frames_size)
        frames[0:3] = [b'a', b'b', b'c']
        frames_void_ptr = spotify.ffi.cast('void *', frames)

        callback = mock.Mock()
        callback.return_value = num_frames
        session = tests.create_real_session(lib_mock)
        session.on('music_delivery', callback)

        result = _SessionCallbacks.music_delivery(
            session._sp_session, sp_audioformat, frames_void_ptr, num_frames)

        callback.assert_called_once_with(
            session, mock.ANY, mock.ANY, num_frames)
        self.assertEqual(
            callback.call_args[0][1]._sp_audioformat, sp_audioformat)
        self.assertEqual(callback.call_args[0][2][:5], b'abc\x00\x00')
        self.assertEqual(result, num_frames)

    def test_music_delivery_without_callback_does_not_consume(self, lib_mock):
        session = tests.create_real_session(lib_mock)

        sp_audioformat = spotify.ffi.new('sp_audioformat *')
        num_frames = 10
        frames = spotify.ffi.new('char[]', 0)
        frames_void_ptr = spotify.ffi.cast('void *', frames)

        result = _SessionCallbacks.music_delivery(
            session._sp_session, sp_audioformat, frames_void_ptr, num_frames)

        self.assertEqual(result, 0)

    def test_play_token_lost_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.PLAY_TOKEN_LOST, callback)

        _SessionCallbacks.play_token_lost(session._sp_session)

        callback.assert_called_once_with(session)

    def test_log_message_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.LOG_MESSAGE, callback)
        data = spotify.ffi.new('char[]', b'a log message\n')

        _SessionCallbacks.log_message(session._sp_session, data)

        callback.assert_called_once_with(session, 'a log message')

    def test_end_of_track_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.END_OF_TRACK, callback)

        _SessionCallbacks.end_of_track(session._sp_session)

        callback.assert_called_once_with(session)

    def test_streaming_error_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.STREAMING_ERROR, callback)

        _SessionCallbacks.streaming_error(
            session._sp_session, int(spotify.ErrorType.NO_STREAM_AVAILABLE))

        callback.assert_called_once_with(
            session, spotify.ErrorType.NO_STREAM_AVAILABLE)

    def test_user_info_updated_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.USER_INFO_UPDATED, callback)

        _SessionCallbacks.user_info_updated(session._sp_session)

        callback.assert_called_once_with(session)

    def test_start_playback_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.START_PLAYBACK, callback)

        _SessionCallbacks.start_playback(session._sp_session)

        callback.assert_called_once_with(session)

    def test_stop_playback_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.STOP_PLAYBACK, callback)

        _SessionCallbacks.stop_playback(session._sp_session)

        callback.assert_called_once_with(session)

    def test_get_audio_buffer_stats_callback(self, lib_mock):
        callback = mock.Mock()
        callback.return_value = spotify.AudioBufferStats(100, 5)
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.GET_AUDIO_BUFFER_STATS, callback)
        sp_audio_buffer_stats = spotify.ffi.new('sp_audio_buffer_stats *')

        _SessionCallbacks.get_audio_buffer_stats(
            session._sp_session, sp_audio_buffer_stats)

        callback.assert_called_once_with(session)
        self.assertEqual(sp_audio_buffer_stats.samples, 100)
        self.assertEqual(sp_audio_buffer_stats.stutter, 5)

    def test_offline_status_updated_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.OFFLINE_STATUS_UPDATED, callback)

        _SessionCallbacks.offline_status_updated(session._sp_session)

        callback.assert_called_once_with(session)

    def test_credentials_blob_updated_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.CREDENTIALS_BLOB_UPDATED, callback)
        data = spotify.ffi.new('char[]', b'a credentials blob')

        _SessionCallbacks.credentials_blob_updated(
            session._sp_session, data)

        callback.assert_called_once_with(session, b'a credentials blob')

    def test_connection_state_updated_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.CONNECTION_STATE_UPDATED, callback)

        _SessionCallbacks.connection_state_updated(session._sp_session)

        callback.assert_called_once_with(session)

    def test_scrobble_error_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.SCROBBLE_ERROR, callback)

        _SessionCallbacks.scrobble_error(
            session._sp_session, int(spotify.ErrorType.LASTFM_AUTH_ERROR))

        callback.assert_called_once_with(
            session, spotify.ErrorType.LASTFM_AUTH_ERROR)

    def test_private_session_mode_changed_callback(self, lib_mock):
        callback = mock.Mock()
        session = tests.create_real_session(lib_mock)
        session.on(spotify.SessionEvent.PRIVATE_SESSION_MODE_CHANGED, callback)

        _SessionCallbacks.private_session_mode_changed(
            session._sp_session, 1)

        callback.assert_called_once_with(session, True)



from __future__ import unicode_literals

import unittest

import spotify
import tests
from tests import mock


@mock.patch('spotify.image.lib', spec=spotify.lib)
class ImageTest(unittest.TestCase):

    def setUp(self):
        self.session = tests.create_session_mock()
        spotify._session_instance = self.session

    def tearDown(self):
        spotify._session_instance = None

    def test_create_without_uri_or_sp_image_fails(self, lib_mock):
        with self.assertRaises(AssertionError):
            spotify.Image(self.session)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_create_from_uri(self, link_mock, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        link_instance_mock = link_mock.return_value
        link_instance_mock.as_image.return_value = spotify.Image(
            self.session, sp_image=sp_image)
        lib_mock.sp_image_create_from_link.return_value = sp_image
        uri = 'spotify:image:foo'

        result = spotify.Image(self.session, uri=uri)

        link_mock.assert_called_with(self.session, uri=uri)
        link_instance_mock.as_image.assert_called_with()
        lib_mock.sp_image_add_ref.assert_called_with(sp_image)
        self.assertEqual(result._sp_image, sp_image)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_create_from_uri_fail_raises_error(self, link_mock, lib_mock):
        link_instance_mock = link_mock.return_value
        link_instance_mock.as_image.return_value = None
        uri = 'spotify:image:foo'

        with self.assertRaises(ValueError):
            spotify.Image(self.session, uri=uri)

    def test_adds_ref_to_sp_image_when_created(self, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)

        spotify.Image(self.session, sp_image=sp_image)

        lib_mock.sp_image_add_ref.assert_called_with(sp_image)

    def test_releases_sp_image_when_image_dies(self, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)

        image = spotify.Image(self.session, sp_image=sp_image)
        image = None  # noqa
        tests.gc_collect()

        lib_mock.sp_image_release.assert_called_with(sp_image)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_repr(self, link_mock, lib_mock):
        link_instance_mock = link_mock.return_value
        link_instance_mock.uri = 'foo'
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        result = repr(image)

        self.assertEqual(result, 'Image(%r)' % 'foo')

    def test_eq(self, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image1 = spotify.Image(self.session, sp_image=sp_image)
        image2 = spotify.Image(self.session, sp_image=sp_image)

        self.assertTrue(image1 == image2)
        self.assertFalse(image1 == 'foo')

    def test_ne(self, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image1 = spotify.Image(self.session, sp_image=sp_image)
        image2 = spotify.Image(self.session, sp_image=sp_image)

        self.assertFalse(image1 != image2)

    def test_hash(self, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image1 = spotify.Image(self.session, sp_image=sp_image)
        image2 = spotify.Image(self.session, sp_image=sp_image)

        self.assertEqual(hash(image1), hash(image2))

    def test_loaded_event_is_unset_by_default(self, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        self.assertFalse(image.loaded_event.is_set())

    def test_create_with_callback(self, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        lib_mock.sp_image_remove_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        lib_mock.sp_image_create.return_value = sp_image
        callback = mock.Mock()

        # Add callback
        image = spotify.Image(
            self.session, sp_image=sp_image, callback=callback)

        lib_mock.sp_image_add_load_callback.assert_called_with(
            sp_image, mock.ANY, mock.ANY)
        image_load_cb = lib_mock.sp_image_add_load_callback.call_args[0][1]
        callback_handle = lib_mock.sp_image_add_load_callback.call_args[0][2]

        # Call calls callback, sets event, and removes callback registration
        self.assertEqual(callback.call_count, 0)
        self.assertEqual(lib_mock.sp_image_remove_load_callback.call_count, 0)
        self.assertFalse(image.loaded_event.is_set())
        image_load_cb(sp_image, callback_handle)
        callback.assert_called_once_with(image)
        lib_mock.sp_image_remove_load_callback.assert_called_with(
            sp_image, image_load_cb, callback_handle)
        self.assertTrue(image.loaded_event.is_set())

    def test_create_with_callback_and_throw_away_image_and_call_load_callback(
            self, lib_mock):

        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        lib_mock.sp_image_remove_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        lib_mock.sp_image_create.return_value = sp_image
        callback = mock.Mock()

        # Add callback
        image = spotify.Image(
            self.session, sp_image=sp_image, callback=callback)
        loaded_event = image.loaded_event

        # Throw away reference to 'image'
        image = None  # noqa
        tests.gc_collect()

        # The mock keeps the handle/userdata alive, thus this test doesn't
        # really test that session._callback_handles keeps the handle alive.

        # Call callback
        image_load_cb = lib_mock.sp_image_add_load_callback.call_args[0][1]
        callback_handle = lib_mock.sp_image_add_load_callback.call_args[0][2]
        image_load_cb(sp_image, callback_handle)

        loaded_event.wait(3)
        self.assertEqual(callback.call_count, 1)
        self.assertEqual(callback.call_args[0][0]._sp_image, sp_image)

    def test_create_with_callback_fails_if_error_adding_callback(
            self, lib_mock):

        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.BAD_API_VERSION)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        callback = mock.Mock()

        with self.assertRaises(spotify.Error):
            spotify.Image(self.session, sp_image=sp_image, callback=callback)

    def test_is_loaded(self, lib_mock):
        lib_mock.sp_image_is_loaded.return_value = 1
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        result = image.is_loaded

        lib_mock.sp_image_is_loaded.assert_called_once_with(sp_image)
        self.assertTrue(result)

    def test_error(self, lib_mock):
        lib_mock.sp_image_error.return_value = int(
            spotify.ErrorType.IS_LOADING)
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        result = image.error

        lib_mock.sp_image_error.assert_called_once_with(sp_image)
        self.assertIs(result, spotify.ErrorType.IS_LOADING)

    @mock.patch('spotify.utils.load')
    def test_load(self, load_mock, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        image.load(10)

        load_mock.assert_called_with(self.session, image, timeout=10)

    def test_format(self, lib_mock):
        lib_mock.sp_image_is_loaded.return_value = 1
        lib_mock.sp_image_format.return_value = int(spotify.ImageFormat.JPEG)
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        result = image.format

        lib_mock.sp_image_format.assert_called_with(sp_image)
        self.assertIs(result, spotify.ImageFormat.JPEG)

    def test_format_is_none_if_unloaded(self, lib_mock):
        lib_mock.sp_image_is_loaded.return_value = 0
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        result = image.format

        lib_mock.sp_image_is_loaded.assert_called_with(sp_image)
        self.assertIsNone(result)

    def test_data(self, lib_mock):
        lib_mock.sp_image_is_loaded.return_value = 1

        size = 20
        data = spotify.ffi.new('char[]', size)
        data[0:3] = [b'a', b'b', b'c']

        def func(sp_image_ptr, data_size_ptr):
            data_size_ptr[0] = size
            return spotify.ffi.cast('void *', data)

        lib_mock.sp_image_data.side_effect = func
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        result = image.data

        lib_mock.sp_image_data.assert_called_with(sp_image, mock.ANY)
        self.assertEqual(result[:5], b'abc\x00\x00')

    def test_data_is_none_if_unloaded(self, lib_mock):
        lib_mock.sp_image_is_loaded.return_value = 0
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        result = image.data

        lib_mock.sp_image_is_loaded.assert_called_with(sp_image)
        self.assertIsNone(result)

    def test_data_uri(self, lib_mock):
        lib_mock.sp_image_format.return_value = int(spotify.ImageFormat.JPEG)
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)

        prop_mock = mock.PropertyMock()
        with mock.patch.object(spotify.Image, 'data', prop_mock):
            image = spotify.Image(self.session, sp_image=sp_image)
            prop_mock.return_value = b'01234\x006789'

            result = image.data_uri

        self.assertEqual(result, 'data:image/jpeg;base64,MDEyMzQANjc4OQ==')

    def test_data_uri_is_none_if_unloaded(self, lib_mock):
        lib_mock.sp_image_is_loaded.return_value = 0
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)

        result = image.data_uri

        self.assertIsNone(result)

    def test_data_uri_fails_if_unknown_image_format(self, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)
        image.__dict__['format'] = mock.Mock(
            return_value=spotify.ImageFormat.UNKNOWN)
        image.__dict__['data'] = mock.Mock(return_value=b'01234\x006789')

        with self.assertRaises(ValueError):
            image.data_uri

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_link_creates_link_to_image(self, link_mock, lib_mock):
        lib_mock.sp_image_add_load_callback.return_value = int(
            spotify.ErrorType.OK)
        sp_image = spotify.ffi.cast('sp_image *', 42)
        image = spotify.Image(self.session, sp_image=sp_image)
        sp_link = spotify.ffi.cast('sp_link *', 43)
        lib_mock.sp_link_create_from_image.return_value = sp_link
        link_mock.return_value = mock.sentinel.link

        result = image.link

        link_mock.assert_called_once_with(
            self.session, sp_link=sp_link, add_ref=False)
        self.assertEqual(result, mock.sentinel.link)


class ImageFormatTest(unittest.TestCase):

    def test_has_constants(self):
        self.assertEqual(spotify.ImageFormat.UNKNOWN, -1)
        self.assertEqual(spotify.ImageFormat.JPEG, 0)


class ImageSizeTest(unittest.TestCase):

    def test_has_size_constants(self):
        self.assertEqual(spotify.ImageSize.NORMAL, 0)
        self.assertEqual(spotify.ImageSize.SMALL, 1)
        self.assertEqual(spotify.ImageSize.LARGE, 2)



from __future__ import unicode_literals

import spotify
from spotify import ffi, lib, utils


__all__ = [
    'ScrobblingState',
    'SocialProvider',
]


class Social(object):

    """Social sharing controller.

    You'll never need to create an instance of this class yourself. You'll find
    it ready to use as the :attr:`~Session.social` attribute on the
    :class:`Session` instance.
    """

    def __init__(self, session):
        self._session = session

    @property
    def private_session(self):
        """Whether the session is private.

        Set to :class:`True` or :class:`False` to change.
        """
        return bool(
            lib.sp_session_is_private_session(self._session._sp_session))

    @private_session.setter
    def private_session(self, value):
        # XXX sp_session_set_private_session() segfaults unless we login and
        # call process_events() at least once before calling it. If we log out
        # again, calling the function still works without segfaults. This bug
        # has been reported to Spotify on IRC.
        if self._session.connection.state != spotify.ConnectionState.LOGGED_IN:
            raise RuntimeError(
                'private_session can only be set when the session is logged '
                'in. This is temporary workaround of a libspotify bug, '
                'causing the application to segfault otherwise.')
        spotify.Error.maybe_raise(lib.sp_session_set_private_session(
            self._session._sp_session, bool(value)))

    def is_scrobbling(self, social_provider):
        """Get the :class:`ScrobblingState` for the given
        ``social_provider``."""
        scrobbling_state = ffi.new('sp_scrobbling_state *')
        spotify.Error.maybe_raise(lib.sp_session_is_scrobbling(
            self._session._sp_session, social_provider, scrobbling_state))
        return spotify.ScrobblingState(scrobbling_state[0])

    def is_scrobbling_possible(self, social_provider):
        """Check if the scrobbling settings should be shown to the user."""
        out = ffi.new('bool *')
        spotify.Error.maybe_raise(lib.sp_session_is_scrobbling_possible(
            self._session._sp_session, social_provider, out))
        return bool(out[0])

    def set_scrobbling(self, social_provider, scrobbling_state):
        """Set the ``scrobbling_state`` for the given ``social_provider``."""
        spotify.Error.maybe_raise(lib.sp_session_set_scrobbling(
            self._session._sp_session, social_provider, scrobbling_state))

    def set_social_credentials(self, social_provider, username, password):
        """Set the user's credentials with a social provider.

        Currently this is only relevant for Last.fm. Call
        :meth:`set_scrobbling` to force an authentication attempt with the
        provider. If authentication fails a
        :attr:`~SessionEvent.SCROBBLE_ERROR` event will be emitted on the
        :class:`Session` object.
        """
        spotify.Error.maybe_raise(lib.sp_session_set_social_credentials(
            self._session._sp_session, social_provider,
            utils.to_char(username), utils.to_char(password)))


@utils.make_enum('SP_SCROBBLING_STATE_')
class ScrobblingState(utils.IntEnum):
    pass


@utils.make_enum('SP_SOCIAL_PROVIDER_')
class SocialProvider(utils.IntEnum):
    pass



from __future__ import unicode_literals

import logging
import weakref

import spotify
from spotify import ffi, lib, serialized, utils
import spotify.connection
import spotify.player
import spotify.social


__all__ = [
    'Session',
    'SessionEvent',
]

logger = logging.getLogger(__name__)


class Session(utils.EventEmitter):

    """The Spotify session.

    If no ``config`` is provided, the default config is used.

    The session object will emit a number of events. See :class:`SessionEvent`
    for a list of all available events and how to connect your own listener
    functions up to get called when the events happens.

    .. warning::

        You can only have one :class:`Session` instance per process. This is a
        libspotify limitation. If you create a second :class:`Session` instance
        in the same process pyspotify will raise a :exc:`RuntimeError` with the
        message "Session has already been initialized".

    :param config: the session config
    :type config: :class:`Config` or :class:`None`
    """

    @serialized
    def __init__(self, config=None):
        super(Session, self).__init__()

        if spotify._session_instance is not None:
            raise RuntimeError('Session has already been initialized')

        if config is not None:
            self.config = config
        else:
            self.config = spotify.Config()

        if self.config.application_key is None:
            self.config.load_application_key_file()

        sp_session_ptr = ffi.new('sp_session **')

        spotify.Error.maybe_raise(lib.sp_session_create(
            self.config._sp_session_config, sp_session_ptr))

        self._sp_session = ffi.gc(sp_session_ptr[0], lib.sp_session_release)

        self._cache = weakref.WeakValueDictionary()
        self._emitters = []
        self._callback_handles = set()

        self.connection = spotify.connection.Connection(self)
        self.offline = spotify.offline.Offline(self)
        self.player = spotify.player.Player(self)
        self.social = spotify.social.Social(self)
        spotify._session_instance = self

    _cache = None
    """A mapping from sp_* objects to their corresponding Python instances.

    The ``_cached`` helper contructors on wrapper objects use this cache for
    finding and returning existing alive wrapper objects for the sp_* object it
    is about to create a wrapper for.

    The cache *does not* keep objects alive. It's only a means for looking up
    the objects if they are kept alive somewhere else in the application.

    Internal attribute.
    """

    _emitters = None
    """A list of event emitters with attached listeners.

    When an event emitter has attached event listeners, we must keep the
    emitter alive for as long as the listeners are attached. This is achieved
    by adding them to this list.

    When creating wrapper objects around sp_* objects we must also return the
    existing wrapper objects instead of creating new ones so that the set of
    event listeners on the wrapper object can be modified. This is achieved
    with a combination of this list and the :attr:`_cache` mapping.

    Internal attribute.
    """

    _callback_handles = None
    """A set of handles returned by :meth:`spotify.ffi.new_handle`.

    These must be kept alive for the handle to remain valid until the callback
    arrives, even if the end user does not maintain a reference to the object
    the callback works on.

    Internal attribute.
    """

    config = None
    """A :class:`Config` instance with the current configuration.

    Once the session has been created, changing the attributes of this object
    will generally have no effect.
    """

    connection = None
    """An :class:`~spotify.connection.Connection` instance for controlling the
    connection to the Spotify servers."""

    offline = None
    """An :class:`~spotify.offline.Offline` instance for controlling offline
    sync."""

    player = None
    """A :class:`~spotify.player.Player` instance for controlling playback."""

    social = None
    """A :class:`~spotify.social.Social` instance for controlling social
    sharing."""

    def login(self, username, password=None, remember_me=False, blob=None):
        """Authenticate to Spotify's servers.

        You can login with one of two combinations:

        - ``username`` and ``password``
        - ``username`` and ``blob``

        To get the ``blob`` string, you must once log in with ``username`` and
        ``password``. You'll then get the ``blob`` string passed to the
        :attr:`~SessionCallbacks.credentials_blob_updated` callback.

        If you set ``remember_me`` to :class:`True`, you can later login to the
        same account without providing any ``username`` or credentials by
        calling :meth:`relogin`.
        """

        username = utils.to_char(username)

        if password is not None:
            password = utils.to_char(password)
            blob = ffi.NULL
        elif blob is not None:
            password = ffi.NULL
            blob = utils.to_char(blob)
        else:
            raise AttributeError('password or blob is required to login')

        spotify.Error.maybe_raise(lib.sp_session_login(
            self._sp_session, username, password, bool(remember_me), blob))

    def logout(self):
        """Log out the current user.

        If you logged in with the ``remember_me`` argument set to
        :class:`True`, you will also need to call :meth:`forget_me` to
        completely remove all credentials of the user that was logged in.
        """
        spotify.Error.maybe_raise(lib.sp_session_logout(self._sp_session))

    @property
    def remembered_user_name(self):
        """The username of the remembered user from a previous :meth:`login`
        call."""
        return utils.get_with_growing_buffer(
            lib.sp_session_remembered_user, self._sp_session)

    def relogin(self):
        """Relogin as the remembered user.

        To be able do this, you must previously have logged in with
        :meth:`login` with the ``remember_me`` argument set to :class:`True`.

        To check what user you'll be logged in as if you call this method, see
        :attr:`remembered_user_name`.
        """
        spotify.Error.maybe_raise(lib.sp_session_relogin(self._sp_session))

    def forget_me(self):
        """Forget the remembered user from a previous :meth:`login` call."""
        spotify.Error.maybe_raise(lib.sp_session_forget_me(self._sp_session))

    @property
    @serialized
    def user(self):
        """The logged in :class:`User`."""
        sp_user = lib.sp_session_user(self._sp_session)
        if sp_user == ffi.NULL:
            return None
        return spotify.User(self, sp_user=sp_user, add_ref=True)

    @property
    @serialized
    def user_name(self):
        """The username of the logged in user."""
        return utils.to_unicode(lib.sp_session_user_name(self._sp_session))

    @property
    @serialized
    def user_country(self):
        """The country of the currently logged in user.

        The :attr:`~SessionEvent.OFFLINE_STATUS_UPDATED` event is emitted on
        the session object when this changes.
        """
        return utils.to_country(lib.sp_session_user_country(self._sp_session))

    @property
    @serialized
    def playlist_container(self):
        """The :class:`PlaylistContainer` for the currently logged in user."""
        sp_playlistcontainer = lib.sp_session_playlistcontainer(
            self._sp_session)
        if sp_playlistcontainer == ffi.NULL:
            return None
        return spotify.PlaylistContainer._cached(
            self, sp_playlistcontainer, add_ref=True)

    @property
    def inbox(self):
        """The inbox :class:`Playlist` for the currently logged in user."""
        sp_playlist = lib.sp_session_inbox_create(self._sp_session)
        if sp_playlist == ffi.NULL:
            return None
        return spotify.Playlist._cached(
            self, sp_playlist=sp_playlist, add_ref=False)

    def set_cache_size(self, size):
        """Set maximum size in MB for libspotify's cache.

        If set to 0 (the default), up to 10% of the free disk space will be
        used."""
        spotify.Error.maybe_raise(lib.sp_session_set_cache_size(
            self._sp_session, size))

    def flush_caches(self):
        """Write all cached data to disk.

        libspotify does this regularly and on logout, so you should never need
        to call this method yourself.
        """
        spotify.Error.maybe_raise(
            lib.sp_session_flush_caches(self._sp_session))

    def preferred_bitrate(self, bitrate):
        """Set preferred :class:`Bitrate` for music streaming."""
        spotify.Error.maybe_raise(lib.sp_session_preferred_bitrate(
            self._sp_session, bitrate))

    def preferred_offline_bitrate(self, bitrate, allow_resync=False):
        """Set preferred :class:`Bitrate` for offline sync.

        If ``allow_resync`` is :class:`True` libspotify may resynchronize
        already synced tracks.
        """
        spotify.Error.maybe_raise(lib.sp_session_preferred_offline_bitrate(
            self._sp_session, bitrate, allow_resync))

    @property
    def volume_normalization(self):
        """Whether volume normalization is active or not.

        Set to :class:`True` or :class:`False` to change.
        """
        return bool(lib.sp_session_get_volume_normalization(self._sp_session))

    @volume_normalization.setter
    def volume_normalization(self, value):
        spotify.Error.maybe_raise(lib.sp_session_set_volume_normalization(
            self._sp_session, value))

    def process_events(self):
        """Process pending events in libspotify.

        This method must be called for most callbacks to be called. Without
        calling this method, you'll only get the callbacks that are called from
        internal libspotify threads. When the
        :attr:`~SessionEvent.NOTIFY_MAIN_THREAD` event is emitted (from an
        internal libspotify thread), it's your job to make sure this method is
        called (from the thread you use for accessing Spotify), so that further
        callbacks can be triggered (from the same thread).

        pyspotify provides an :class:`~spotify.EventLoop` that you can use for
        processing events when needed.
        """
        next_timeout = ffi.new('int *')

        spotify.Error.maybe_raise(lib.sp_session_process_events(
            self._sp_session, next_timeout))

        return next_timeout[0]

    def inbox_post_tracks(
            self, canonical_username, tracks, message, callback=None):
        """Post a ``message`` and one or more ``tracks`` to the inbox of the
        user with the given ``canonical_username``.

        ``tracks`` can be a single :class:`~spotify.Track` or a list of
        :class:`~spotify.Track` objects.

        Returns an :class:`InboxPostResult` that can be used to check if the
        request completed successfully.

        If callback isn't :class:`None`, it is called with an
        :class:`InboxPostResult` instance when the request has completed.
        """
        return spotify.InboxPostResult(
            self, canonical_username, tracks, message, callback)

    def get_starred(self, canonical_username=None):
        """Get the starred :class:`Playlist` for the user with
        ``canonical_username``.

        If ``canonical_username`` isn't specified, the starred playlist for
        the currently logged in user is returned.
        """
        if canonical_username is None:
            sp_playlist = lib.sp_session_starred_create(self._sp_session)
        else:
            sp_playlist = lib.sp_session_starred_for_user_create(
                self._sp_session, utils.to_bytes(canonical_username))
        if sp_playlist == ffi.NULL:
            return None
        return spotify.Playlist._cached(self, sp_playlist, add_ref=False)

    def get_published_playlists(self, canonical_username=None):
        """Get the :class:`PlaylistContainer` of published playlists for the
        user with ``canonical_username``.

        If ``canonical_username`` isn't specified, the published container for
        the currently logged in user is returned.
        """
        if canonical_username is None:
            canonical_username = ffi.NULL
        else:
            canonical_username = utils.to_bytes(canonical_username)
        sp_playlistcontainer = (
            lib.sp_session_publishedcontainer_for_user_create(
                self._sp_session, canonical_username))
        if sp_playlistcontainer == ffi.NULL:
            return None
        return spotify.PlaylistContainer._cached(
            self, sp_playlistcontainer, add_ref=False)

    def get_link(self, uri):
        """
        Get :class:`Link` from any Spotify URI.

        A link can be created from a string containing a Spotify URI on the
        form ``spotify:...``.

        Example::

            >>> session = spotify.Session()
            # ...
            >>> session.get_link(
            ...     'spotify:track:2Foc5Q5nqNiosCNqttzHof')
            Link('spotify:track:2Foc5Q5nqNiosCNqttzHof')
            >>> session.get_link(
            ...     'http://open.spotify.com/track/4wl1dK5dHGp3Ig51stvxb0')
            Link('spotify:track:4wl1dK5dHGp3Ig51stvxb0')
        """
        return spotify.Link(self, uri=uri)

    def get_track(self, uri):
        """
        Get :class:`Track` from a Spotify track URI.

        Example::

            >>> session = spotify.Session()
            # ...
            >>> track = session.get_track(
            ...     'spotify:track:2Foc5Q5nqNiosCNqttzHof')
            >>> track.load().name
            u'Get Lucky'
        """
        return spotify.Track(self, uri=uri)

    def get_local_track(
            self, artist=None, title=None, album=None, length=None):
        """
        Get :class:`Track` for a local track.

        Spotify's official clients supports adding your local music files to
        Spotify so they can be played in the Spotify client. These are not
        synced with Spotify's servers or between your devices and there is not
        trace of them in your Spotify user account. The exception is when you
        add one of these local tracks to a playlist or mark them as starred.
        This creates a "local track" which pyspotify also will be able to
        observe.

        "Local tracks" can be recognized in several ways:

        - The track's URI will be of the form
          ``spotify:local:ARTIST:ALBUM:TITLE:LENGTH_IN_SECONDS``. Any of the
          parts in all caps can be left out if there is no information
          available. That is, ``spotify:local::::`` is a valid local track URI.

        - :attr:`Link.type` will be :class:`LinkType.LOCALTRACK` for the
          track's link.

        - :attr:`Track.is_local` will be :class:`True` for the track.

        This method can be used to create local tracks that can be starred or
        added to playlists.

        ``artist`` may be an artist name. ``title`` may be a track name.
        ``album`` may be an album name. ``length`` may be a track length in
        milliseconds.

        Note that when creating a local track you provide the length in
        milliseconds, while the local track URI contains the length in seconds.
        """

        if artist is None:
            artist = ''
        if title is None:
            title = ''
        if album is None:
            album = ''
        if length is None:
            length = -1

        artist = utils.to_char(artist)
        title = utils.to_char(title)
        album = utils.to_char(album)
        sp_track = lib.sp_localtrack_create(artist, title, album, length)

        return spotify.Track(self, sp_track=sp_track, add_ref=False)

    def get_album(self, uri):
        """
        Get :class:`Album` from a Spotify album URI.

        Example::

            >>> session = spotify.Session()
            # ...
            >>> album = session.get_album(
            ...     'spotify:album:6wXDbHLesy6zWqQawAa91d')
            >>> album.load().name
            u'Forward / Return'
        """
        return spotify.Album(self, uri=uri)

    def get_artist(self, uri):
        """
        Get :class:`Artist` from a Spotify artist URI.

        Example::

            >>> session = spotify.Session()
            # ...
            >>> artist = session.get_artist(
            ...     'spotify:artist:22xRIphSN7IkPVbErICu7s')
            >>> artist.load().name
            u'Rob Dougan'
        """
        return spotify.Artist(self, uri=uri)

    def get_playlist(self, uri):
        """
        Get :class:`Playlist` from a Spotify playlist URI.

        Example::

            >>> session = spotify.Session()
            # ...
            >>> playlist = session.get_playlist(
            ...     'spotify:user:fiat500c:playlist:54k50VZdvtnIPt4d8RBCmZ')
            >>> playlist.load().name
            u'500C feelgood playlist'
        """
        return spotify.Playlist(self, uri=uri)

    def get_user(self, uri):
        """
        Get :class:`User` from a Spotify user URI.

        Example::

            >>> session = spotify.Session()
            # ...
            >>> user = session.get_user('spotify:user:jodal')
            >>> user.load().display_name
            u'jodal'
        """
        return spotify.User(self, uri=uri)

    def get_image(self, uri, callback=None):
        """
        Get :class:`Image` from a Spotify image URI.

        If ``callback`` isn't :class:`None`, it is expected to be a callable
        that accepts a single argument, an :class:`Image` instance, when
        the image is done loading.

        Example::

            >>> session = spotify.Session()
            # ...
            >>> image = session.get_image(
            ...     'spotify:image:a0bdcbe11b5cd126968e519b5ed1050b0e8183d0')
            >>> image.load().data_uri[:50]
            u'data:image/jpeg;base64,/9j/4AAQSkZJRgABAQEBLAEsAAD'
        """
        return spotify.Image(self, uri=uri, callback=callback)

    def search(
            self, query, callback=None,
            track_offset=0, track_count=20,
            album_offset=0, album_count=20,
            artist_offset=0, artist_count=20,
            playlist_offset=0, playlist_count=20,
            search_type=None):
        """
        Search Spotify for tracks, albums, artists, and playlists matching
        ``query``.

        The ``query`` string can be free format, or use some prefixes like
        ``title:`` and ``artist:`` to limit what to match on. There is no
        official docs on the search query format, but there's a `Spotify blog
        post
        <https://www.spotify.com/blog/archives/2008/01/22/searching-spotify/>`_
        from 2008 with some examples.

        If ``callback`` isn't :class:`None`, it is expected to be a callable
        that accepts a single argument, a :class:`Search` instance, when
        the search completes.

        The ``*_offset`` and ``*_count`` arguments can be used to retrieve more
        search results. libspotify will currently not respect ``*_count``
        values higher than 200, though this may change at any time as the limit
        isn't documented in any official docs. If you want to retrieve more
        than 200 results, you'll have to search multiple times with different
        ``*_offset`` values. See the ``*_total`` attributes on the
        :class:`Search` to see how many results exists, and to figure out
        how many searches you'll need to make to retrieve everything.

        ``search_type`` is a :class:`SearchType` value. It defaults to
        :attr:`SearchType.STANDARD`.

        Returns a :class:`Search` instance.
        """
        return spotify.Search(
            self, query=query, callback=callback,
            track_offset=track_offset, track_count=track_count,
            album_offset=album_offset, album_count=album_count,
            artist_offset=artist_offset, artist_count=artist_count,
            playlist_offset=playlist_offset, playlist_count=playlist_count,
            search_type=search_type)

    def get_toplist(
            self, type=None, region=None, canonical_username=None,
            callback=None):
        """Get a :class:`Toplist` of artists, albums, or tracks that are the
        currently most popular worldwide or in a specific region.

        ``type`` is a :class:`ToplistType` instance that specifies the type of
        toplist to create.

        ``region`` is either a :class:`ToplistRegion` instance, or a 2-letter
        ISO 3166-1 country code as a unicode string, that specifies the
        geographical region to create a toplist for.

        If ``region`` is :attr:`ToplistRegion.USER` and ``canonical_username``
        isn't specified, the region of the current user will be used. If
        ``canonical_username`` is specified, the region of the specified user
        will be used instead.

        If ``callback`` isn't :class:`None`, it is expected to be a callable
        that accepts a single argument, a :class:`Toplist` instance, when the
        toplist request completes.

        Example::

            >>> import spotify
            >>> session = spotify.Session()
            # ...
            >>> toplist = session.get_toplist(
            ...     type=spotify.ToplistType.TRACKS, region='US')
            >>> toplist.load()
            >>> len(toplist.tracks)
            100
            >>> len(toplist.artists)
            0
            >>> toplist.tracks[0]
            Track(u'spotify:track:2dLLR6qlu5UJ5gk0dKz0h3')
        """
        return spotify.Toplist(
            self, type=type, region=region,
            canonical_username=canonical_username, callback=callback)


class SessionEvent(object):

    """Session events.

    Using the :class:`Session` object, you can register listener functions to
    be called when various session related events occurs. This class enumerates
    the available events and the arguments your listener functions will be
    called with.

    Example usage::

        import spotify

        def logged_in(session, error_type):
            if error_type is spotify.ErrorType.OK:
                print('Logged in as %s' % session.user)
            else:
                print('Login failed: %s' % error_type)

        session = spotify.Session()
        session.on(spotify.SessionEvent.LOGGED_IN, logged_in)
        session.login('alice', 's3cret')

    All events will cause debug log statements to be emitted, even if no
    listeners are registered. Thus, there is no need to register listener
    functions just to log that they're called.
    """

    LOGGED_IN = 'logged_in'
    """Called when login has completed.

    Note that even if login has succeeded, that does not mean that you're
    online yet as libspotify may have cached enough information to let you
    authenticate with Spotify while offline.

    This event should be used to get notified about login errors. To get
    notified about the authentication and connection state, refer to the
    :attr:`SessionEvent.CONNECTION_STATE_UPDATED` event.

    :param session: the current session
    :type session: :class:`Session`
    :param error_type: the login error type
    :type error_type: :class:`ErrorType`
    """

    LOGGED_OUT = 'logged_out'
    """Called when logout has completed or there is a permanent connection
    error.

    :param session: the current session
    :type session: :class:`Session`
    """

    METADATA_UPDATED = 'metadata_updated'
    """Called when some metadata has been updated.

    There is no way to know what metadata was updated, so you'll have to
    refresh all you metadata caches.

    :param session: the current session
    :type session: :class:`Session`
    """

    CONNECTION_ERROR = 'connection_error'
    """Called when there is a connection error and libspotify has problems
    reconnecting to the Spotify service.

    May be called repeatedly as long as the problem persists. Will be called
    with an :attr:`ErrorType.OK` error when the problem is resolved.

    :param session: the current session
    :type session: :class:`Session`
    :param error_type: the connection error type
    :type error_type: :class:`ErrorType`
    """

    MESSAGE_TO_USER = 'message_to_user'
    """Called when libspotify wants to show a message to the end user.

    :param session: the current session
    :type session: :class:`Session`
    :param data: the message
    :type data: text
    """

    NOTIFY_MAIN_THREAD = 'notify_main_thread'
    """Called when processing on the main thread is needed.

    When this is called, you should call :meth:`~Session.process_events` from
    your main thread. Failure to do so may cause request timeouts, or a lost
    connection.

    .. warning::

        This event is emitted from an internal libspotify thread. Thus, your
        event listener must not block, and must use proper synchronization
        around anything it does.

    :param session: the current session
    :type session: :class:`Session`
    """

    MUSIC_DELIVERY = 'music_delivery'
    """Called when there is decompressed audio data available.

    If the function returns a lower number of frames consumed than
    ``num_frames``, libspotify will retry delivery of the unconsumed frames in
    about 100ms. This can be used for rate limiting if libspotify is giving you
    audio data too fast.

    .. note::

        You can register at most one event listener for this event.

    .. warning::

        This event is emitted from an internal libspotify thread. Thus, your
        event listener must not block, and must use proper synchronization
        around anything it does.

    :param session: the current session
    :type session: :class:`Session`
    :param audio_format: the audio format
    :type audio_format: :class:`AudioFormat`
    :param frames: the audio frames
    :type frames: bytestring
    :param num_frames: the number of frames
    :type num_frames: int
    :returns: the number of frames consumed
    """

    PLAY_TOKEN_LOST = 'play_token_lost'
    """Music has been paused because an account only allows music to be played
    from one location simultaneously.

    When this event is emitted, you should pause playback.

    :param session: the current session
    :type session: :class:`Session`
    """

    LOG_MESSAGE = 'log_message'
    """Called when libspotify have something to log.

    Note that pyspotify logs this for you, so you'll probably never need to
    register a listener for this event.

    :param session: the current session
    :type session: :class:`Session`
    :param data: the message
    :type data: text
    """

    END_OF_TRACK = 'end_of_track'
    """Called when all audio data for the current track has been delivered.

    :param session: the current session
    :type session: :class:`Session`
    """

    STREAMING_ERROR = 'streaming_error'
    """Called when audio streaming cannot start or continue.

    :param session: the current session
    :type session: :class:`Session`
    :param error_type: the streaming error type
    :type error_type: :class:`ErrorType`
    """

    USER_INFO_UPDATED = 'user_info_updated'
    """Called when anything related to :class:`User` objects is updated.

    :param session: the current session
    :type session: :class:`Session`
    """

    START_PLAYBACK = 'start_playback'
    """Called when audio playback should start.

    You need to implement a listener for the :attr:`GET_AUDIO_BUFFER_STATS`
    event for the :attr:`START_PLAYBACK` event to be useful.

    .. warning::

        This event is emitted from an internal libspotify thread. Thus, your
        event listener must not block, and must use proper synchronization
        around anything it does.

    :param session: the current session
    :type session: :class:`Session`
    """

    STOP_PLAYBACK = 'stop_playback'
    """Called when audio playback should stop.

    You need to implement a listener for the :attr:`GET_AUDIO_BUFFER_STATS`
    event for the :attr:`STOP_PLAYBACK` event to be useful.

    .. warning::

        This event is emitted from an internal libspotify thread. Thus, your
        event listener must not block, and must use proper synchronization
        around anything it does.

    :param session: the current session
    :type session: :class:`Session`
    """

    GET_AUDIO_BUFFER_STATS = 'get_audio_buffer_stats'
    """Called to query the application about its audio buffer.

    .. note::

        You can register at most one event listener for this event.

    .. warning::

        This event is emitted from an internal libspotify thread. Thus, your
        event listener must not block, and must use proper synchronization
        around anything it does.

    :param session: the current session
    :type session: :class:`Session`
    :returns: an :class:`AudioBufferStats` instance
    """

    OFFLINE_STATUS_UPDATED = 'offline_status_updated'
    """Called when offline sync status is updated.

    :param session: the current session
    :type session: :class:`Session`
    """

    CREDENTIALS_BLOB_UPDATED = 'credentials_blob_updated'
    """Called when storable credentials have been updated, typically right
    after login.

    The ``blob`` argument can be stored and later passed to
    :meth:`~Session.login` to login without storing the user's password.

    :param session: the current session
    :type session: :class:`Session`
    :param blob: the authentication blob
    :type blob: bytestring
    """

    CONNECTION_STATE_UPDATED = 'connection_state_updated'
    """Called when the connection state is updated.

    The connection state includes login, logout, offline mode, etc.

    :param session: the current session
    :type session: :class:`Session`
    """

    SCROBBLE_ERROR = 'scrobble_error'
    """Called when there is a scrobble error event.

    :param session: the current session
    :type session: :class:`Session`
    :param error_type: the scrobble error type
    :type error_type: :class:`ErrorType`
    """

    PRIVATE_SESSION_MODE_CHANGED = 'private_session_mode_changed'
    """Called when there is a change in the private session mode.

    :param session: the current session
    :type session: :class:`Session`
    :param is_private: whether the session is private
    :type is_private: bool
    """


class _SessionCallbacks(object):

    """Internal class."""

    @classmethod
    def get_struct(cls):
        return ffi.new('sp_session_callbacks *', {
            'logged_in': cls.logged_in,
            'logged_out': cls.logged_out,
            'metadata_updated': cls.metadata_updated,
            'connection_error': cls.connection_error,
            'message_to_user': cls.message_to_user,
            'notify_main_thread': cls.notify_main_thread,
            'music_delivery': cls.music_delivery,
            'play_token_lost': cls.play_token_lost,
            'log_message': cls.log_message,
            'end_of_track': cls.end_of_track,
            'streaming_error': cls.streaming_error,
            'userinfo_updated': cls.user_info_updated,
            'start_playback': cls.start_playback,
            'stop_playback': cls.stop_playback,
            'get_audio_buffer_stats': cls.get_audio_buffer_stats,
            'offline_status_updated': cls.offline_status_updated,
            'credentials_blob_updated': cls.credentials_blob_updated,
            'connectionstate_updated': cls.connection_state_updated,
            'scrobble_error': cls.scrobble_error,
            'private_session_mode_changed': cls.private_session_mode_changed,
        })

    # XXX Avoid use of the spotify._session_instance global in the following
    # callbacks.

    @staticmethod
    @ffi.callback('void(sp_session *, sp_error)')
    def logged_in(sp_session, sp_error):
        if not spotify._session_instance:
            return
        error_type = spotify.ErrorType(sp_error)
        if error_type == spotify.ErrorType.OK:
            logger.info('Spotify logged in')
        else:
            logger.error('Spotify login error: %r', error_type)
        spotify._session_instance.emit(
            SessionEvent.LOGGED_IN, spotify._session_instance, error_type)

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def logged_out(sp_session):
        if not spotify._session_instance:
            return
        logger.info('Spotify logged out')
        spotify._session_instance.emit(
            SessionEvent.LOGGED_OUT, spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def metadata_updated(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('Metadata updated')
        spotify._session_instance.emit(
            SessionEvent.METADATA_UPDATED, spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *, sp_error)')
    def connection_error(sp_session, sp_error):
        if not spotify._session_instance:
            return
        error_type = spotify.ErrorType(sp_error)
        logger.error('Spotify connection error: %r', error_type)
        spotify._session_instance.emit(
            SessionEvent.CONNECTION_ERROR,
            spotify._session_instance, error_type)

    @staticmethod
    @ffi.callback('void(sp_session *, const char *)')
    def message_to_user(sp_session, data):
        if not spotify._session_instance:
            return
        data = utils.to_unicode(data).strip()
        logger.debug('Message to user: %s', data)
        spotify._session_instance.emit(
            SessionEvent.MESSAGE_TO_USER, spotify._session_instance, data)

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def notify_main_thread(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('Notify main thread')
        spotify._session_instance.emit(
            SessionEvent.NOTIFY_MAIN_THREAD, spotify._session_instance)

    @staticmethod
    @ffi.callback(
        'int(sp_session *, const sp_audioformat *, const void *, int)')
    def music_delivery(sp_session, sp_audioformat, frames, num_frames):
        if not spotify._session_instance:
            return 0
        if spotify._session_instance.num_listeners(
                SessionEvent.MUSIC_DELIVERY) == 0:
            logger.debug('Music delivery, but no listener')
            return 0
        audio_format = spotify.AudioFormat(sp_audioformat)
        frames_buffer = ffi.buffer(
            frames, audio_format.frame_size() * num_frames)
        frames_bytes = frames_buffer[:]
        num_frames_consumed = spotify._session_instance.call(
            SessionEvent.MUSIC_DELIVERY,
            spotify._session_instance, audio_format, frames_bytes, num_frames)
        logger.debug(
            'Music delivery of %d frames, %d consumed', num_frames,
            num_frames_consumed)
        return num_frames_consumed

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def play_token_lost(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('Play token lost')
        spotify._session_instance.emit(
            SessionEvent.PLAY_TOKEN_LOST, spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *, const char *)')
    def log_message(sp_session, data):
        if not spotify._session_instance:
            return
        data = utils.to_unicode(data).strip()
        logger.debug('libspotify log message: %s', data)
        spotify._session_instance.emit(
            SessionEvent.LOG_MESSAGE, spotify._session_instance, data)

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def end_of_track(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('End of track')
        spotify._session_instance.emit(
            SessionEvent.END_OF_TRACK, spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *, sp_error)')
    def streaming_error(sp_session, sp_error):
        if not spotify._session_instance:
            return
        error_type = spotify.ErrorType(sp_error)
        logger.error('Spotify streaming error: %r', error_type)
        spotify._session_instance.emit(
            SessionEvent.STREAMING_ERROR,
            spotify._session_instance, error_type)

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def user_info_updated(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('User info updated')
        spotify._session_instance.emit(
            SessionEvent.USER_INFO_UPDATED, spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def start_playback(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('Start playback called')
        spotify._session_instance.emit(
            SessionEvent.START_PLAYBACK, spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def stop_playback(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('Stop playback called')
        spotify._session_instance.emit(
            SessionEvent.STOP_PLAYBACK, spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *, sp_audio_buffer_stats *)')
    def get_audio_buffer_stats(sp_session, sp_audio_buffer_stats):
        if not spotify._session_instance:
            return
        if spotify._session_instance.num_listeners(
                SessionEvent.GET_AUDIO_BUFFER_STATS) == 0:
            logger.debug('Audio buffer stats requested, but no listener')
            return
        logger.debug('Audio buffer stats requested')
        stats = spotify._session_instance.call(
            SessionEvent.GET_AUDIO_BUFFER_STATS, spotify._session_instance)
        sp_audio_buffer_stats.samples = stats.samples
        sp_audio_buffer_stats.stutter = stats.stutter

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def offline_status_updated(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('Offline status updated')
        spotify._session_instance.emit(
            SessionEvent.OFFLINE_STATUS_UPDATED, spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *, const char *)')
    def credentials_blob_updated(sp_session, data):
        if not spotify._session_instance:
            return
        data = ffi.string(data)
        logger.debug('Credentials blob updated: %r', data)
        spotify._session_instance.emit(
            SessionEvent.CREDENTIALS_BLOB_UPDATED,
            spotify._session_instance, data)

    @staticmethod
    @ffi.callback('void(sp_session *)')
    def connection_state_updated(sp_session):
        if not spotify._session_instance:
            return
        logger.debug('Connection state updated')
        spotify._session_instance.emit(
            SessionEvent.CONNECTION_STATE_UPDATED,
            spotify._session_instance)

    @staticmethod
    @ffi.callback('void(sp_session *, sp_error)')
    def scrobble_error(sp_session, sp_error):
        if not spotify._session_instance:
            return
        error_type = spotify.ErrorType(sp_error)
        logger.error('Spotify scrobble error: %r', error_type)
        spotify._session_instance.emit(
            SessionEvent.SCROBBLE_ERROR,
            spotify._session_instance, error_type)

    @staticmethod
    @ffi.callback('void(sp_session *, bool)')
    def private_session_mode_changed(sp_session, is_private):
        if not spotify._session_instance:
            return
        is_private = bool(is_private)
        status = 'private' if is_private else 'public'
        logger.debug('Private session mode changed: %s', status)
        spotify._session_instance.emit(
            SessionEvent.PRIVATE_SESSION_MODE_CHANGED,
            spotify._session_instance, is_private)



from __future__ import unicode_literals

import time
import unittest

import spotify
from spotify.utils import load
import tests
from tests import mock


class Foo(object):

    def __init__(self, session):
        self._session = session

    @property
    def is_loaded(self):
        return True

    def load(self, timeout=None):
        return load(self._session, self, timeout=timeout)


class FooWithError(Foo):

    @property
    def error(self):
        return spotify.Error(spotify.Error.OK)


@mock.patch('spotify.utils.time')
@mock.patch.object(Foo, 'is_loaded', new_callable=mock.PropertyMock)
class LoadableTest(unittest.TestCase):

    def setUp(self):
        self.session = tests.create_session_mock()
        self.session.connection.state = spotify.ConnectionState.LOGGED_IN

    def test_load_raises_error_if_not_logged_in(
            self, is_loaded_mock, time_mock):
        is_loaded_mock.return_value = False
        self.session.connection.state = spotify.ConnectionState.LOGGED_OUT
        foo = Foo(self.session)

        with self.assertRaises(spotify.Error):
            foo.load()

    def test_load_raises_error_if_offline(
            self, is_loaded_mock, time_mock):
        is_loaded_mock.return_value = False
        self.session.connection.state = spotify.ConnectionState.OFFLINE
        foo = Foo(self.session)

        with self.assertRaises(spotify.Error):
            foo.load()

    def test_load_returns_immediately_if_offline_but_already_loaded(
            self, is_loaded_mock, time_mock):
        is_loaded_mock.return_value = True
        self.session.connection.state = spotify.ConnectionState.OFFLINE
        foo = Foo(self.session)

        result = foo.load()

        self.assertEqual(result, foo)
        self.assertEqual(self.session.process_events.call_count, 0)

    def test_load_raises_error_when_timeout_is_reached(
            self, is_loaded_mock, time_mock):
        is_loaded_mock.return_value = False
        time_mock.time.side_effect = time.time
        foo = Foo(self.session)

        with self.assertRaises(spotify.Timeout):
            foo.load(timeout=0)

    def test_load_processes_events_until_loaded(
            self, is_loaded_mock, time_mock):
        is_loaded_mock.side_effect = [False, False, False, False, False, True]
        time_mock.time.side_effect = time.time

        foo = Foo(self.session)
        foo.load()

        self.assertEqual(self.session.process_events.call_count, 2)
        self.assertEqual(time_mock.sleep.call_count, 2)

    @mock.patch.object(FooWithError, 'error', new_callable=mock.PropertyMock)
    def test_load_raises_exception_on_error(
            self, error_mock, is_loaded_mock, time_mock):
        error_mock.side_effect = [
            spotify.ErrorType.IS_LOADING, spotify.ErrorType.OTHER_PERMANENT]
        is_loaded_mock.side_effect = [False, False, True]

        foo = FooWithError(self.session)

        with self.assertRaises(spotify.Error):
            foo.load()

        self.assertEqual(self.session.process_events.call_count, 1)
        self.assertEqual(time_mock.sleep.call_count, 0)

    def test_load_raises_exception_on_error_even_if_already_loaded(
            self, is_loaded_mock, time_mock):
        is_loaded_mock.return_value = True

        foo = Foo(self.session)
        foo.error = spotify.ErrorType.OTHER_PERMANENT

        with self.assertRaises(spotify.Error):
            foo.load()

    def test_load_does_not_abort_on_is_loading_error(
            self, is_loaded_mock, time_mock):
        is_loaded_mock.side_effect = [False, False, False, False, False, True]
        time_mock.time.side_effect = time.time

        foo = Foo(self.session)
        foo.error = spotify.ErrorType.IS_LOADING
        foo.load()

        self.assertEqual(self.session.process_events.call_count, 2)
        self.assertEqual(time_mock.sleep.call_count, 2)

    def test_load_returns_self(self, is_loaded_mock, time_mock):
        is_loaded_mock.return_value = True

        foo = Foo(self.session)
        result = foo.load()

        self.assertEqual(result, foo)



from __future__ import unicode_literals

import logging

import spotify
from spotify import ffi, lib, serialized, utils


__all__ = [
    'PlaylistTrack',
]

logger = logging.getLogger(__name__)


class PlaylistTrack(object):

    """A playlist track with metadata specific to the playlist.

    Use :attr:`~spotify.Playlist.tracks_with_metadata` to get a list of
    :class:`PlaylistTrack`.
    """

    def __init__(self, session, sp_playlist, index):
        self._session = session

        lib.sp_playlist_add_ref(sp_playlist)
        self._sp_playlist = ffi.gc(sp_playlist, lib.sp_playlist_release)

        self._index = index

    def __repr__(self):
        return 'PlaylistTrack(uri=%r, creator=%r, create_time=%d)' % (
            self.track.link.uri, self.creator, self.create_time)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self._sp_playlist == other._sp_playlist and
                self._index == other._index)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self._sp_playlist, self._index))

    @property
    @serialized
    def track(self):
        """The :class:`~spotify.Track`."""
        return spotify.Track(
            self._session,
            sp_track=lib.sp_playlist_track(self._sp_playlist, self._index),
            add_ref=True)

    @property
    def create_time(self):
        """When the track was added to the playlist, as seconds since Unix
        epoch.
        """
        return lib.sp_playlist_track_create_time(
            self._sp_playlist, self._index)

    @property
    @serialized
    def creator(self):
        """The :class:`~spotify.User` that added the track to the playlist."""
        return spotify.User(
            self._session,
            sp_user=lib.sp_playlist_track_creator(
                self._sp_playlist, self._index),
            add_ref=True)

    def is_seen(self):
        return bool(lib.sp_playlist_track_seen(self._sp_playlist, self._index))

    def set_seen(self, value):
        spotify.Error.maybe_raise(lib.sp_playlist_track_set_seen(
            self._sp_playlist, self._index, int(value)))

    seen = property(is_seen, set_seen)
    """Whether the track is marked as seen or not."""

    @property
    @serialized
    def message(self):
        """A message attached to the track. Typically used in the inbox."""
        message = lib.sp_playlist_track_message(self._sp_playlist, self._index)
        return utils.to_unicode_or_none(message)



from __future__ import unicode_literals

import logging
import threading

import spotify
from spotify import ffi, lib, serialized, utils


__all__ = [
    'InboxPostResult',
]

logger = logging.getLogger(__name__)


class InboxPostResult(object):

    """The result object returned by :meth:`Session.inbox_post_tracks`."""

    @serialized
    def __init__(
            self, session, canonical_username=None, tracks=None, message='',
            callback=None, sp_inbox=None, add_ref=True):

        assert canonical_username and tracks or sp_inbox, \
            'canonical_username and tracks, or sp_inbox, is required'

        self._session = session
        self.loaded_event = threading.Event()

        if sp_inbox is None:
            canonical_username = utils.to_char(canonical_username)

            if isinstance(tracks, spotify.Track):
                tracks = [tracks]

            message = utils.to_char(message)

            handle = ffi.new_handle((self._session, self, callback))
            self._session._callback_handles.add(handle)

            sp_inbox = lib.sp_inbox_post_tracks(
                self._session._sp_session, canonical_username,
                [t._sp_track for t in tracks], len(tracks),
                message, _inboxpost_complete_callback, handle)
            add_ref = True

            if sp_inbox == ffi.NULL:
                raise spotify.Error('Inbox post request failed to initialize')

        if add_ref:
            lib.sp_inbox_add_ref(sp_inbox)
        self._sp_inbox = ffi.gc(sp_inbox, lib.sp_inbox_release)

    loaded_event = None
    """:class:`threading.Event` that is set when the inbox post result is
    loaded.
    """

    def __repr__(self):
        if not self.loaded_event.is_set():
            return 'InboxPostResult(<pending>)'
        else:
            return 'InboxPostResult(%s)' % self.error._name

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._sp_inbox == other._sp_inbox
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._sp_inbox)

    @property
    def error(self):
        """An :class:`ErrorType` associated with the inbox post result.

        Check to see if there was problems posting to the inbox.
        """
        return spotify.ErrorType(lib.sp_inbox_error(self._sp_inbox))


@ffi.callback('void(sp_inbox *, void *)')
@serialized
def _inboxpost_complete_callback(sp_inbox, handle):
    logger.debug('inboxpost_complete_callback called')
    if handle == ffi.NULL:
        logger.warning(
            'pyspotify inboxpost_complete_callback called without userdata')
        return
    (session, inbox_post_result, callback) = ffi.from_handle(handle)
    session._callback_handles.remove(handle)
    inbox_post_result.loaded_event.set()
    if callback is not None:
        callback(inbox_post_result)



from __future__ import unicode_literals

import unittest

import spotify
from spotify import utils
import tests
from tests import mock


@mock.patch('spotify.artist.lib', spec=spotify.lib)
class ArtistTest(unittest.TestCase):

    def setUp(self):
        self.session = tests.create_session_mock()

    def test_create_without_uri_or_sp_artist_fails(self, lib_mock):
        with self.assertRaises(AssertionError):
            spotify.Artist(self.session)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_create_from_uri(self, link_mock, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        link_instance_mock = link_mock.return_value
        link_instance_mock.as_artist.return_value = spotify.Artist(
            self.session, sp_artist=sp_artist)
        lib_mock.sp_link_as_artist.return_value = sp_artist
        uri = 'spotify:artist:foo'

        result = spotify.Artist(self.session, uri=uri)

        link_mock.assert_called_with(self.session, uri=uri)
        link_instance_mock.as_artist.assert_called_with()
        lib_mock.sp_artist_add_ref.assert_called_with(sp_artist)
        self.assertEqual(result._sp_artist, sp_artist)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_create_from_uri_fail_raises_error(self, link_mock, lib_mock):
        link_instance_mock = link_mock.return_value
        link_instance_mock.as_artist.return_value = None
        lib_mock.sp_link_as_artist.return_value = spotify.ffi.NULL
        uri = 'spotify:artist:foo'

        with self.assertRaises(ValueError):
            spotify.Artist(self.session, uri=uri)

    def test_adds_ref_to_sp_artist_when_created(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)

        spotify.Artist(self.session, sp_artist=sp_artist)

        lib_mock.sp_artist_add_ref.assert_called_with(sp_artist)

    def test_releases_sp_artist_when_artist_dies(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)

        artist = spotify.Artist(self.session, sp_artist=sp_artist)
        artist = None  # noqa
        tests.gc_collect()

        lib_mock.sp_artist_release.assert_called_with(sp_artist)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_repr(self, link_mock, lib_mock):
        link_instance_mock = link_mock.return_value
        link_instance_mock.uri = 'foo'
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)

        result = repr(artist)

        self.assertEqual(result, 'Artist(%r)' % 'foo')

    def test_eq(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist1 = spotify.Artist(self.session, sp_artist=sp_artist)
        artist2 = spotify.Artist(self.session, sp_artist=sp_artist)

        self.assertTrue(artist1 == artist2)
        self.assertFalse(artist1 == 'foo')

    def test_ne(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist1 = spotify.Artist(self.session, sp_artist=sp_artist)
        artist2 = spotify.Artist(self.session, sp_artist=sp_artist)

        self.assertFalse(artist1 != artist2)

    def test_hash(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist1 = spotify.Artist(self.session, sp_artist=sp_artist)
        artist2 = spotify.Artist(self.session, sp_artist=sp_artist)

        self.assertEqual(hash(artist1), hash(artist2))

    def test_name(self, lib_mock):
        lib_mock.sp_artist_name.return_value = spotify.ffi.new(
            'char[]', b'Foo Bar Baz')
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)

        result = artist.name

        lib_mock.sp_artist_name.assert_called_once_with(sp_artist)
        self.assertEqual(result, 'Foo Bar Baz')

    def test_name_is_none_if_unloaded(self, lib_mock):
        lib_mock.sp_artist_name.return_value = spotify.ffi.new('char[]', b'')
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)

        result = artist.name

        lib_mock.sp_artist_name.assert_called_once_with(sp_artist)
        self.assertIsNone(result)

    def test_is_loaded(self, lib_mock):
        lib_mock.sp_artist_is_loaded.return_value = 1
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)

        result = artist.is_loaded

        lib_mock.sp_artist_is_loaded.assert_called_once_with(sp_artist)
        self.assertTrue(result)

    @mock.patch('spotify.utils.load')
    def test_load(self, load_mock, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)

        artist.load(10)

        load_mock.assert_called_with(self.session, artist, timeout=10)

    @mock.patch('spotify.Image', spec=spotify.Image)
    def test_portrait(self, image_mock, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)
        sp_image_id = spotify.ffi.new('char[]', b'portrait-id')
        lib_mock.sp_artist_portrait.return_value = sp_image_id
        sp_image = spotify.ffi.cast('sp_image *', 43)
        lib_mock.sp_image_create.return_value = sp_image
        image_mock.return_value = mock.sentinel.image
        image_size = spotify.ImageSize.SMALL
        callback = mock.Mock()

        result = artist.portrait(image_size, callback=callback)

        self.assertIs(result, mock.sentinel.image)
        lib_mock.sp_artist_portrait.assert_called_with(
            sp_artist, int(image_size))
        lib_mock.sp_image_create.assert_called_with(
            self.session._sp_session, sp_image_id)

        # Since we *created* the sp_image, we already have a refcount of 1 and
        # shouldn't increase the refcount when wrapping this sp_image in an
        # Image object
        image_mock.assert_called_with(
            self.session, sp_image=sp_image, add_ref=False, callback=callback)

    @mock.patch('spotify.Image', spec=spotify.Image)
    def test_portrait_defaults_to_normal_size(self, image_mock, lib_mock):
        sp_image_id = spotify.ffi.new('char[]', b'portrait-id')
        lib_mock.sp_artist_portrait.return_value = sp_image_id
        sp_image = spotify.ffi.cast('sp_image *', 43)
        lib_mock.sp_image_create.return_value = sp_image
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)

        artist.portrait()

        lib_mock.sp_artist_portrait.assert_called_with(
            sp_artist, int(spotify.ImageSize.NORMAL))

    def test_portrait_is_none_if_null(self, lib_mock):
        lib_mock.sp_artist_portrait.return_value = spotify.ffi.NULL
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)

        result = artist.portrait()

        lib_mock.sp_artist_portrait.assert_called_with(
            sp_artist, int(spotify.ImageSize.NORMAL))
        self.assertIsNone(result)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_portrait_link_creates_link_to_portrait(self, link_mock, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)
        sp_link = spotify.ffi.cast('sp_link *', 43)
        lib_mock.sp_link_create_from_artist_portrait.return_value = sp_link
        link_mock.return_value = mock.sentinel.link
        image_size = spotify.ImageSize.SMALL

        result = artist.portrait_link(image_size)

        lib_mock.sp_link_create_from_artist_portrait.assert_called_once_with(
            sp_artist, int(image_size))
        link_mock.assert_called_once_with(
            self.session, sp_link=sp_link, add_ref=False)
        self.assertEqual(result, mock.sentinel.link)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_portrait_link_defaults_to_normal_size(self, link_mock, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)
        sp_link = spotify.ffi.cast('sp_link *', 43)
        lib_mock.sp_link_create_from_artist_portrait.return_value = sp_link
        link_mock.return_value = mock.sentinel.link

        artist.portrait_link()

        lib_mock.sp_link_create_from_artist_portrait.assert_called_once_with(
            sp_artist, int(spotify.ImageSize.NORMAL))

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_link_creates_link_to_artist(self, link_mock, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)
        sp_link = spotify.ffi.cast('sp_link *', 43)
        lib_mock.sp_link_create_from_artist.return_value = sp_link
        link_mock.return_value = mock.sentinel.link

        result = artist.link

        link_mock.assert_called_once_with(
            self.session, sp_link=sp_link, add_ref=False)
        self.assertEqual(result, mock.sentinel.link)


@mock.patch('spotify.artist.lib', spec=spotify.lib)
class ArtistBrowserTest(unittest.TestCase):

    def setUp(self):
        self.session = tests.create_session_mock()
        spotify._session_instance = self.session

    def tearDown(self):
        spotify._session_instance = None

    def test_create_without_artist_or_sp_artistbrowse_fails(self, lib_mock):
        with self.assertRaises(AssertionError):
            spotify.ArtistBrowser(self.session)

    def test_create_from_artist(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 43)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        lib_mock.sp_artistbrowse_create.return_value = sp_artistbrowse

        result = artist.browse()

        lib_mock.sp_artistbrowse_create.assert_called_with(
            self.session._sp_session, sp_artist,
            int(spotify.ArtistBrowserType.FULL), mock.ANY, mock.ANY)
        self.assertIsInstance(result, spotify.ArtistBrowser)

        artistbrowse_complete_cb = (
            lib_mock.sp_artistbrowse_create.call_args[0][3])
        userdata = lib_mock.sp_artistbrowse_create.call_args[0][4]
        self.assertFalse(result.loaded_event.is_set())
        artistbrowse_complete_cb(sp_artistbrowse, userdata)
        self.assertTrue(result.loaded_event.is_set())

    def test_create_from_artist_with_type_and_callback(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 43)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        lib_mock.sp_artistbrowse_create.return_value = sp_artistbrowse
        callback = mock.Mock()

        result = artist.browse(
            type=spotify.ArtistBrowserType.NO_TRACKS, callback=callback)

        lib_mock.sp_artistbrowse_create.assert_called_with(
            self.session._sp_session, sp_artist,
            int(spotify.ArtistBrowserType.NO_TRACKS), mock.ANY, mock.ANY)
        artistbrowse_complete_cb = (
            lib_mock.sp_artistbrowse_create.call_args[0][3])
        userdata = lib_mock.sp_artistbrowse_create.call_args[0][4]
        artistbrowse_complete_cb(sp_artistbrowse, userdata)

        result.loaded_event.wait(3)
        callback.assert_called_with(result)

    def test_browser_is_gone_before_callback_is_called(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 43)
        artist = spotify.Artist(self.session, sp_artist=sp_artist)
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        lib_mock.sp_artistbrowse_create.return_value = sp_artistbrowse
        callback = mock.Mock()

        result = spotify.ArtistBrowser(
            self.session, artist=artist, callback=callback)
        loaded_event = result.loaded_event
        result = None  # noqa
        tests.gc_collect()

        # The mock keeps the handle/userdata alive, thus this test doesn't
        # really test that session._callback_handles keeps the handle alive.
        artistbrowse_complete_cb = (
            lib_mock.sp_artistbrowse_create.call_args[0][3])
        userdata = lib_mock.sp_artistbrowse_create.call_args[0][4]
        artistbrowse_complete_cb(sp_artistbrowse, userdata)

        loaded_event.wait(3)
        self.assertEqual(callback.call_count, 1)
        self.assertEqual(
            callback.call_args[0][0]._sp_artistbrowse, sp_artistbrowse)

    def test_adds_ref_to_sp_artistbrowse_when_created(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)

        spotify.ArtistBrowser(self.session, sp_artistbrowse=sp_artistbrowse)

        lib_mock.sp_artistbrowse_add_ref.assert_called_with(sp_artistbrowse)

    def test_releases_sp_artistbrowse_when_artist_dies(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)

        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        browser = None  # noqa
        tests.gc_collect()

        lib_mock.sp_artistbrowse_release.assert_called_with(sp_artistbrowse)

    @mock.patch('spotify.Link', spec=spotify.Link)
    def test_repr(self, link_mock, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        lib_mock.sp_artistbrowse_is_loaded.return_value = 1
        sp_artist = spotify.ffi.cast('sp_artist *', 42)
        lib_mock.sp_artistbrowse_artist.return_value = sp_artist
        link_instance_mock = link_mock.return_value
        link_instance_mock.uri = 'foo'

        result = repr(browser)

        self.assertEqual(result, 'ArtistBrowser(%r)' % 'foo')

    def test_repr_if_unloaded(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        lib_mock.sp_artistbrowse_is_loaded.return_value = 0

        result = repr(browser)

        self.assertEqual(result, 'ArtistBrowser(<not loaded>)')

    def test_eq(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser1 = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        browser2 = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        self.assertTrue(browser1 == browser2)
        self.assertFalse(browser1 == 'foo')

    def test_ne(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser1 = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        browser2 = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        self.assertFalse(browser1 != browser2)

    def test_hash(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser1 = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        browser2 = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        self.assertEqual(hash(browser1), hash(browser2))

    def test_is_loaded(self, lib_mock):
        lib_mock.sp_artistbrowse_is_loaded.return_value = 1
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.is_loaded

        lib_mock.sp_artistbrowse_is_loaded.assert_called_once_with(
            sp_artistbrowse)
        self.assertTrue(result)

    @mock.patch('spotify.utils.load')
    def test_load(self, load_mock, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        browser.load(10)

        load_mock.assert_called_with(self.session, browser, timeout=10)

    def test_error(self, lib_mock):
        lib_mock.sp_artistbrowse_error.return_value = int(
            spotify.ErrorType.OTHER_PERMANENT)
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.error

        lib_mock.sp_artistbrowse_error.assert_called_once_with(sp_artistbrowse)
        self.assertIs(result, spotify.ErrorType.OTHER_PERMANENT)

    def test_backend_request_duration(self, lib_mock):
        lib_mock.sp_artistbrowse_backend_request_duration.return_value = 137
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.backend_request_duration

        lib_mock.sp_artistbrowse_backend_request_duration.assert_called_with(
            sp_artistbrowse)
        self.assertEqual(result, 137)

    def test_backend_request_duration_when_not_loaded(self, lib_mock):
        lib_mock.sp_artistbrowse_is_loaded.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.backend_request_duration

        lib_mock.sp_artistbrowse_is_loaded.assert_called_with(sp_artistbrowse)
        self.assertEqual(
            lib_mock.sp_artistbrowse_backend_request_duration.call_count, 0)
        self.assertIsNone(result)

    def test_artist(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        sp_artist = spotify.ffi.cast('sp_artist *', 43)
        lib_mock.sp_artistbrowse_artist.return_value = sp_artist

        result = browser.artist

        self.assertIsInstance(result, spotify.Artist)
        self.assertEqual(result._sp_artist, sp_artist)

    def test_artist_when_not_loaded(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        lib_mock.sp_artistbrowse_artist.return_value = spotify.ffi.NULL

        result = browser.artist

        lib_mock.sp_artistbrowse_artist.assert_called_with(sp_artistbrowse)
        self.assertIsNone(result)

    @mock.patch('spotify.Image', spec=spotify.Image)
    def test_portraits(self, image_mock, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        lib_mock.sp_artistbrowse_num_portraits.return_value = 1
        image_id = spotify.ffi.new('char[]', b'image-id')
        lib_mock.sp_artistbrowse_portrait.return_value = image_id
        sp_image = spotify.ffi.cast('sp_image *', 43)
        lib_mock.sp_image_create.return_value = sp_image
        image_mock.return_value = mock.sentinel.image
        callback = mock.Mock()

        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 1)
        result = browser.portraits(callback=callback)
        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 2)

        self.assertEqual(len(result), 1)
        lib_mock.sp_artistbrowse_num_portraits.assert_called_with(
            sp_artistbrowse)

        item = result[0]
        self.assertIs(item, mock.sentinel.image)
        self.assertEqual(lib_mock.sp_artistbrowse_portrait.call_count, 1)
        lib_mock.sp_artistbrowse_portrait.assert_called_with(
            sp_artistbrowse, 0)

        # Since we *created* the sp_image, we already have a refcount of 1 and
        # shouldn't increase the refcount when wrapping this sp_image in an
        # Image object
        image_mock.assert_called_with(
            self.session, sp_image=sp_image, add_ref=False, callback=callback)

    def test_portraits_if_no_portraits(self, lib_mock):
        lib_mock.sp_artistbrowse_num_portraits.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.portraits()

        self.assertEqual(len(result), 0)
        lib_mock.sp_artistbrowse_num_portraits.assert_called_with(
            sp_artistbrowse)
        self.assertEqual(lib_mock.sp_artistbrowse_portrait.call_count, 0)

    def test_portraits_if_unloaded(self, lib_mock):
        lib_mock.sp_artistbrowse_is_loaded.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.portraits()

        lib_mock.sp_artistbrowse_is_loaded.assert_called_with(sp_artistbrowse)
        self.assertEqual(len(result), 0)

    @mock.patch('spotify.track.lib', spec=spotify.lib)
    def test_tracks(self, track_lib_mock, lib_mock):
        sp_track = spotify.ffi.cast('sp_track *', 43)
        lib_mock.sp_artistbrowse_num_tracks.return_value = 1
        lib_mock.sp_artistbrowse_track.return_value = sp_track
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 1)
        result = browser.tracks
        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 2)

        self.assertEqual(len(result), 1)
        lib_mock.sp_artistbrowse_num_tracks.assert_called_with(sp_artistbrowse)

        item = result[0]
        self.assertIsInstance(item, spotify.Track)
        self.assertEqual(item._sp_track, sp_track)
        self.assertEqual(lib_mock.sp_artistbrowse_track.call_count, 1)
        lib_mock.sp_artistbrowse_track.assert_called_with(sp_artistbrowse, 0)
        track_lib_mock.sp_track_add_ref.assert_called_with(sp_track)

    def test_tracks_if_no_tracks(self, lib_mock):
        lib_mock.sp_artistbrowse_num_tracks.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.tracks

        self.assertEqual(len(result), 0)
        lib_mock.sp_artistbrowse_num_tracks.assert_called_with(sp_artistbrowse)
        self.assertEqual(lib_mock.sp_artistbrowse_track.call_count, 0)

    def test_tracks_if_unloaded(self, lib_mock):
        lib_mock.sp_artistbrowse_is_loaded.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.tracks

        lib_mock.sp_artistbrowse_is_loaded.assert_called_with(sp_artistbrowse)
        self.assertEqual(len(result), 0)

    @mock.patch('spotify.track.lib', spec=spotify.lib)
    def test_tophit_tracks(self, track_lib_mock, lib_mock):
        sp_track = spotify.ffi.cast('sp_track *', 43)
        lib_mock.sp_artistbrowse_num_tophit_tracks.return_value = 1
        lib_mock.sp_artistbrowse_tophit_track.return_value = sp_track
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 1)
        result = browser.tophit_tracks
        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 2)

        self.assertEqual(len(result), 1)
        lib_mock.sp_artistbrowse_num_tophit_tracks.assert_called_with(
            sp_artistbrowse)

        item = result[0]
        self.assertIsInstance(item, spotify.Track)
        self.assertEqual(item._sp_track, sp_track)
        self.assertEqual(lib_mock.sp_artistbrowse_tophit_track.call_count, 1)
        lib_mock.sp_artistbrowse_tophit_track.assert_called_with(
            sp_artistbrowse, 0)
        track_lib_mock.sp_track_add_ref.assert_called_with(sp_track)

    def test_tophit_tracks_if_no_tracks(self, lib_mock):
        lib_mock.sp_artistbrowse_num_tophit_tracks.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.tophit_tracks

        self.assertEqual(len(result), 0)
        lib_mock.sp_artistbrowse_num_tophit_tracks.assert_called_with(
            sp_artistbrowse)
        self.assertEqual(lib_mock.sp_artistbrowse_track.call_count, 0)

    def test_tophit_tracks_if_unloaded(self, lib_mock):
        lib_mock.sp_artistbrowse_is_loaded.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.tophit_tracks

        lib_mock.sp_artistbrowse_is_loaded.assert_called_with(sp_artistbrowse)
        self.assertEqual(len(result), 0)

    @mock.patch('spotify.album.lib', spec=spotify.lib)
    def test_albums(self, album_lib_mock, lib_mock):
        sp_album = spotify.ffi.cast('sp_album *', 43)
        lib_mock.sp_artistbrowse_num_albums.return_value = 1
        lib_mock.sp_artistbrowse_album.return_value = sp_album
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 1)
        result = browser.albums
        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 2)

        self.assertEqual(len(result), 1)
        lib_mock.sp_artistbrowse_num_albums.assert_called_with(sp_artistbrowse)

        item = result[0]
        self.assertIsInstance(item, spotify.Album)
        self.assertEqual(item._sp_album, sp_album)
        self.assertEqual(lib_mock.sp_artistbrowse_album.call_count, 1)
        lib_mock.sp_artistbrowse_album.assert_called_with(sp_artistbrowse, 0)
        album_lib_mock.sp_album_add_ref.assert_called_with(sp_album)

    def test_albums_if_no_albums(self, lib_mock):
        lib_mock.sp_artistbrowse_num_albums.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.albums

        self.assertEqual(len(result), 0)
        lib_mock.sp_artistbrowse_num_albums.assert_called_with(sp_artistbrowse)
        self.assertEqual(lib_mock.sp_artistbrowse_album.call_count, 0)

    def test_albums_if_unloaded(self, lib_mock):
        lib_mock.sp_artistbrowse_is_loaded.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.albums

        lib_mock.sp_artistbrowse_is_loaded.assert_called_with(sp_artistbrowse)
        self.assertEqual(len(result), 0)

    def test_similar_artists(self, lib_mock):
        sp_artist = spotify.ffi.cast('sp_artist *', 43)
        lib_mock.sp_artistbrowse_num_similar_artists.return_value = 1
        lib_mock.sp_artistbrowse_similar_artist.return_value = sp_artist
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 1)
        result = browser.similar_artists
        self.assertEqual(lib_mock.sp_artistbrowse_add_ref.call_count, 2)

        self.assertEqual(len(result), 1)
        lib_mock.sp_artistbrowse_num_similar_artists.assert_called_with(
            sp_artistbrowse)

        item = result[0]
        self.assertIsInstance(item, spotify.Artist)
        self.assertEqual(item._sp_artist, sp_artist)
        self.assertEqual(
            lib_mock.sp_artistbrowse_similar_artist.call_count, 1)
        lib_mock.sp_artistbrowse_similar_artist.assert_called_with(
            sp_artistbrowse, 0)
        lib_mock.sp_artist_add_ref.assert_called_with(sp_artist)

    def test_similar_artists_if_no_artists(self, lib_mock):
        lib_mock.sp_artistbrowse_num_similar_artists.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.similar_artists

        self.assertEqual(len(result), 0)
        lib_mock.sp_artistbrowse_num_similar_artists.assert_called_with(
            sp_artistbrowse)
        self.assertEqual(lib_mock.sp_artistbrowse_similar_artist.call_count, 0)

    def test_similar_artists_if_unloaded(self, lib_mock):
        lib_mock.sp_artistbrowse_is_loaded.return_value = 0
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)

        result = browser.similar_artists

        lib_mock.sp_artistbrowse_is_loaded.assert_called_with(sp_artistbrowse)
        self.assertEqual(len(result), 0)

    def test_biography(self, lib_mock):
        sp_artistbrowse = spotify.ffi.cast('sp_artistbrowse *', 42)
        browser = spotify.ArtistBrowser(
            self.session, sp_artistbrowse=sp_artistbrowse)
        biography = spotify.ffi.new('char[]', b'Lived, played, and died')
        lib_mock.sp_artistbrowse_biography.return_value = biography

        result = browser.biography

        self.assertIsInstance(result, utils.text_type)
        self.assertEqual(result, 'Lived, played, and died')


class ArtistBrowserTypeTest(unittest.TestCase):

    def test_has_constants(self):
        self.assertEqual(spotify.ArtistBrowserType.FULL, 0)
        self.assertEqual(spotify.ArtistBrowserType.NO_TRACKS, 1)
        self.assertEqual(spotify.ArtistBrowserType.NO_ALBUMS, 2)
